#  Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time
from unittest import mock

import ddt
from nova import exception
from os_win import exceptions as os_win_exc
from oslo_utils import fileutils
from six.moves import builtins

from compute_hyperv.nova import constants
from compute_hyperv.nova import pathutils
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class PathUtilsTestCase(test_base.HyperVBaseTestCase):
    """Unit tests for the Hyper-V PathUtils class."""

    def setUp(self):
        super(PathUtilsTestCase, self).setUp()
        self.fake_instance_dir = os.path.join('C:', 'fake_instance_dir')
        self.fake_instance_name = 'fake_instance_name'

        self._pathutils = pathutils.PathUtils()

    @mock.patch.object(pathutils.PathUtils, 'copy')
    @mock.patch.object(os.path, 'isfile')
    @mock.patch.object(os, 'listdir')
    def test_copy_folder_files(self, mock_listdir, mock_isfile, mock_copy):
        src_dir = 'src'
        dest_dir = 'dest'
        fname = 'tmp_file.txt'
        subdir = 'tmp_folder'
        src_fname = os.path.join(src_dir, fname)
        dest_fname = os.path.join(dest_dir, fname)

        # making sure src_subdir is not copied.
        mock_listdir.return_value = [fname, subdir]
        mock_isfile.side_effect = [True, False]

        self._pathutils.copy_folder_files(src_dir, dest_dir)
        mock_copy.assert_called_once_with(src_fname, dest_fname)

    @ddt.data({'conf_instances_path': r'c:\inst_dir',
               'expected_dir': r'c:\inst_dir'},
              {'conf_instances_path': r'c:\inst_dir',
               'remote_server': 'fake_remote',
               'expected_dir': r'\\fake_remote\c$\inst_dir'},
              {'conf_instances_path': r'\\fake_share\fake_path',
               'remote_server': 'fake_remote',
               'expected_dir': r'\\fake_share\fake_path'},
              {'conf_instances_path_share': r'inst_share',
               'remote_server': 'fake_remote',
               'expected_dir': r'\\fake_remote\inst_share'})
    @ddt.unpack
    def test_get_instances_dir(self, expected_dir, remote_server=None,
                               conf_instances_path='',
                               conf_instances_path_share=''):
        self.flags(instances_path=conf_instances_path)
        self.flags(instances_path_share=conf_instances_path_share,
                   group='hyperv')

        instances_dir = self._pathutils.get_instances_dir(remote_server)

        self.assertEqual(expected_dir, instances_dir)

    def test_get_remote_path_share(self):
        fake_remote_path = '\\\\fake_path'

        actual_path = self._pathutils.get_remote_path(mock.sentinel.server,
                                                      fake_remote_path)
        self.assertEqual(fake_remote_path, actual_path)

    @mock.patch.object(pathutils.os, 'getenv')
    def test_get_remote_path_csv(self, mock_getenv):
        mock_getenv.return_value = 'C:'
        fake_server = 'fake_server'
        fake_remote_path = 'C:\\ClusterStorage\\Volume1\\fake_dir'

        actual_path = self._pathutils.get_remote_path(fake_server,
                                                      fake_remote_path)

        self.assertEqual(fake_remote_path, actual_path)
        mock_getenv.assert_called_once_with('SYSTEMDRIVE', 'C:')

    def test_get_remote_path_normal(self):
        fake_server = 'fake_server'
        fake_remote_path = 'C:\\fake_path'

        actual_path = self._pathutils.get_remote_path(fake_server,
                                                      fake_remote_path)

        expected_path = ('\\\\%(remote_server)s\\%(path)s' %
                         dict(remote_server=fake_server,
                              path=fake_remote_path.replace(':', '$')))
        self.assertEqual(expected_path, actual_path)

    @mock.patch.object(pathutils.PathUtils, 'get_instances_dir')
    @mock.patch.object(pathutils.PathUtils, 'check_dir')
    def test_get_instances_sub_dir(self, mock_check_dir,
                                   mock_get_instances_dir):
        fake_instances_dir = 'fake_instances_dir'
        mock_get_instances_dir.return_value = fake_instances_dir

        sub_dir = 'fake_subdir'
        expected_path = os.path.join(fake_instances_dir, sub_dir)

        path = self._pathutils._get_instances_sub_dir(
            sub_dir,
            remote_server=mock.sentinel.remote_server,
            create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

        self.assertEqual(expected_path, path)

        mock_get_instances_dir.assert_called_once_with(
            mock.sentinel.remote_server)
        mock_check_dir.assert_called_once_with(
            expected_path,
            create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

    @ddt.data({'create_dir': True, 'remove_dir': False},
              {'create_dir': False, 'remove_dir': True})
    @ddt.unpack
    @mock.patch.object(pathutils.PathUtils, 'check_create_dir')
    @mock.patch.object(pathutils.PathUtils, 'check_remove_dir')
    def test_check_dir(self, mock_check_remove_dir, mock_check_create_dir,
                       create_dir, remove_dir):
        self._pathutils.check_dir(
            mock.sentinel.dir, create_dir=create_dir, remove_dir=remove_dir)

        if create_dir:
            mock_check_create_dir.assert_called_once_with(mock.sentinel.dir)
        else:
            self.assertFalse(mock_check_create_dir.called)

        if remove_dir:
            mock_check_remove_dir.assert_called_once_with(mock.sentinel.dir)
        else:
            self.assertFalse(mock_check_remove_dir.called)

    @mock.patch.object(pathutils.PathUtils, 'check_create_dir')
    def test_check_dir_exc(self, mock_check_create_dir):

        class FakeWindowsError(Exception):
            def __init__(self, winerror=None):
                self.winerror = winerror

        mock_check_create_dir.side_effect = FakeWindowsError(
            pathutils.ERROR_INVALID_NAME)
        with mock.patch.object(builtins, 'WindowsError',
                               FakeWindowsError, create=True):
            self.assertRaises(exception.AdminRequired,
                              self._pathutils.check_dir,
                              mock.sentinel.dir_name,
                              create_dir=True)

    @mock.patch.object(pathutils.PathUtils, 'check_dir')
    def test_get_instance_migr_revert_dir(self, mock_check_dir):
        dir_name = 'fake_dir'
        expected_dir_name = '%s_revert' % dir_name

        revert_dir = self._pathutils.get_instance_migr_revert_dir(
            dir_name, create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

        self.assertEqual(expected_dir_name, revert_dir)
        mock_check_dir.assert_called_once_with(expected_dir_name,
                                               mock.sentinel.create_dir,
                                               mock.sentinel.remove_dir)

    @ddt.data({},
              {'configured_dir_exists': True},
              {'vm_exists': True},
              {'vm_exists': True,
               'remote_server': mock.sentinel.remote_server})
    @ddt.unpack
    @mock.patch.object(pathutils.PathUtils, '_get_instances_sub_dir')
    @mock.patch.object(pathutils.PathUtils, 'get_remote_path')
    @mock.patch.object(pathutils.PathUtils, 'check_dir')
    @mock.patch.object(pathutils.os.path, 'exists')
    @mock.patch('os_win.utilsfactory.get_vmutils')
    def test_get_instance_dir(self, mock_get_vmutils,
                              mock_exists,
                              mock_check_dir,
                              mock_get_remote_path,
                              mock_get_instances_sub_dir,
                              configured_dir_exists=False,
                              remote_server=None, vm_exists=False):
        mock_get_instances_sub_dir.return_value = mock.sentinel.configured_dir
        mock_exists.return_value = configured_dir_exists

        expected_vmutils = (self._pathutils._vmutils
                            if not remote_server
                            else mock_get_vmutils.return_value)
        mock_get_root_dir = expected_vmutils.get_vm_config_root_dir
        mock_get_root_dir.side_effect = (
            (mock.sentinel.config_root_dir,)
            if vm_exists
            else os_win_exc.HyperVVMNotFoundException(
                vm_name=mock.sentinel.instance_name))

        mock_get_remote_path.return_value = mock.sentinel.remote_root_dir

        instance_dir = self._pathutils.get_instance_dir(
            mock.sentinel.instance_name,
            remote_server=remote_server,
            create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

        if configured_dir_exists or not vm_exists:
            expected_instance_dir = mock.sentinel.configured_dir
        else:
            # In this case, we expect the instance location to be
            # retrieved from the vm itself.
            mock_get_root_dir.assert_called_once_with(
                mock.sentinel.instance_name)

            if remote_server:
                expected_instance_dir = mock.sentinel.remote_root_dir
                mock_get_remote_path.assert_called_once_with(
                    mock.sentinel.remote_server,
                    mock.sentinel.config_root_dir)
            else:
                expected_instance_dir = mock.sentinel.config_root_dir

        self.assertEqual(expected_instance_dir, instance_dir)

        mock_get_instances_sub_dir.assert_called_once_with(
            mock.sentinel.instance_name, remote_server,
            create_dir=False, remove_dir=False)
        mock_check_dir.assert_called_once_with(
            expected_instance_dir,
            create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

    def _mock_lookup_configdrive_path(self, ext, rescue=False):
        self._pathutils.get_instance_dir = mock.MagicMock(
            return_value=self.fake_instance_dir)

        def mock_exists(*args, **kwargs):
            path = args[0]
            return True if path[(path.rfind('.') + 1):] == ext else False
        self._pathutils.exists = mock_exists
        configdrive_path = self._pathutils.lookup_configdrive_path(
            self.fake_instance_name, rescue)
        return configdrive_path

    def _test_lookup_configdrive_path(self, rescue=False):
        configdrive_name = 'configdrive'
        if rescue:
            configdrive_name += '-rescue'

        for format_ext in constants.DISK_FORMAT_MAP:
            configdrive_path = self._mock_lookup_configdrive_path(format_ext,
                                                                  rescue)
            expected_path = os.path.join(self.fake_instance_dir,
                                         configdrive_name + '.' + format_ext)
            self.assertEqual(expected_path, configdrive_path)

    def test_lookup_configdrive_path(self):
        self._test_lookup_configdrive_path()

    def test_lookup_rescue_configdrive_path(self):
        self._test_lookup_configdrive_path(rescue=True)

    def test_lookup_configdrive_path_non_exist(self):
        self._pathutils.get_instance_dir = mock.MagicMock(
            return_value=self.fake_instance_dir)
        self._pathutils.exists = mock.MagicMock(return_value=False)
        configdrive_path = self._pathutils.lookup_configdrive_path(
            self.fake_instance_name)
        self.assertIsNone(configdrive_path)

    @mock.patch.object(pathutils.PathUtils, 'check_dir')
    @mock.patch.object(pathutils.PathUtils, 'get_instance_dir')
    def test_export_dir(self, mock_get_instance_dir, mock_check_dir):
        mock_get_instance_dir.return_value = self.fake_instance_dir

        export_dir = self._pathutils.get_export_dir(
            mock.sentinel.instance_name, create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

        expected_dir = os.path.join(self.fake_instance_dir, 'export')
        self.assertEqual(expected_dir, export_dir)
        mock_get_instance_dir.assert_called_once_with(
            mock.sentinel.instance_name, create_dir=mock.sentinel.create_dir)
        mock_check_dir.assert_called_once_with(
            expected_dir, create_dir=mock.sentinel.create_dir,
            remove_dir=mock.sentinel.remove_dir)

    def test_copy_vm_console_logs(self):
        fake_local_logs = [mock.sentinel.log_path,
                           mock.sentinel.archived_log_path]
        fake_remote_logs = [mock.sentinel.remote_log_path,
                            mock.sentinel.remote_archived_log_path]

        self._pathutils.exists = mock.Mock(return_value=True)
        self._pathutils.copy = mock.Mock()
        self._pathutils.get_vm_console_log_paths = mock.Mock(
            side_effect=[fake_local_logs, fake_remote_logs])

        self._pathutils.copy_vm_console_logs(mock.sentinel.instance_name,
                                            mock.sentinel.dest_host)

        self._pathutils.get_vm_console_log_paths.assert_has_calls(
            [mock.call(mock.sentinel.instance_name),
             mock.call(mock.sentinel.instance_name,
                       remote_server=mock.sentinel.dest_host)])
        self._pathutils.copy.assert_has_calls([
            mock.call(mock.sentinel.log_path,
                      mock.sentinel.remote_log_path),
            mock.call(mock.sentinel.archived_log_path,
                      mock.sentinel.remote_archived_log_path)])

    @mock.patch.object(pathutils.PathUtils, 'get_base_vhd_dir')
    @mock.patch.object(pathutils.PathUtils, 'exists')
    def _test_get_image_path(self, mock_exists, mock_get_base_vhd_dir,
                             found=True):
        fake_image_name = 'fake_image_name'
        if found:
            mock_exists.side_effect = [False, True]
            expected_path = os.path.join('fake_base_dir',
                                         'fake_image_name.vhdx')
        else:
            mock_exists.return_value = False
            expected_path = None
        mock_get_base_vhd_dir.return_value = 'fake_base_dir'

        res = self._pathutils.get_image_path(fake_image_name)

        mock_get_base_vhd_dir.assert_called_once_with()
        self.assertEqual(expected_path, res)

    def test_get_image_path(self):
        self._test_get_image_path()

    def test_get_image_path_not_found(self):
        self._test_get_image_path(found=False)

    @mock.patch('os.path.getmtime')
    @mock.patch.object(pathutils, 'time')
    def test_get_age_of_file(self, mock_time, mock_getmtime):
        mock_time.time.return_value = time.time()
        mock_getmtime.return_value = mock_time.time.return_value - 42

        actual_age = self._pathutils.get_age_of_file(mock.sentinel.filename)
        self.assertEqual(42, actual_age)
        mock_time.time.assert_called_once_with()
        mock_getmtime.assert_called_once_with(mock.sentinel.filename)

    @mock.patch('os.path.exists')
    @mock.patch('tempfile.NamedTemporaryFile')
    def test_check_dirs_shared_storage(self, mock_named_tempfile,
                                       mock_exists):
        fake_src_dir = 'fake_src_dir'
        fake_dest_dir = 'fake_dest_dir'

        mock_exists.return_value = True
        mock_tmpfile = mock_named_tempfile.return_value.__enter__.return_value
        mock_tmpfile.name = 'fake_tmp_fname'
        expected_src_tmp_path = os.path.join(fake_src_dir,
                                             mock_tmpfile.name)

        self._pathutils.check_dirs_shared_storage(
            fake_src_dir, fake_dest_dir)

        mock_named_tempfile.assert_called_once_with(dir=fake_dest_dir)
        mock_exists.assert_called_once_with(expected_src_tmp_path)

    @mock.patch.object(pathutils.PathUtils, 'check_dirs_shared_storage')
    @mock.patch.object(pathutils.PathUtils, 'get_instances_dir')
    def test_check_remote_instances_shared(self, mock_get_instances_dir,
                                           mock_check_dirs_shared_storage):
        mock_get_instances_dir.side_effect = [mock.sentinel.local_inst_dir,
                                              mock.sentinel.remote_inst_dir]

        shared_storage = self._pathutils.check_remote_instances_dir_shared(
            mock.sentinel.dest)

        self.assertEqual(mock_check_dirs_shared_storage.return_value,
                         shared_storage)
        mock_get_instances_dir.assert_has_calls(
            [mock.call(), mock.call(mock.sentinel.dest)])
        mock_check_dirs_shared_storage.assert_called_once_with(
            mock.sentinel.local_inst_dir, mock.sentinel.remote_inst_dir)

    @mock.patch.object(os, 'close')
    @mock.patch('tempfile.mkstemp')
    @mock.patch.object(pathutils.PathUtils, 'get_instance_dir')
    def test_check_instance_shared_storage_local(self, mock_get_instance_dir,
                                                 mock_mkstemp, mock_close):
        mock_instance = mock.Mock()
        mock_mkstemp.return_value = (mock.sentinel.tmp_fd,
                                     mock.sentinel.tmp_file)

        ret_val = self._pathutils.check_instance_shared_storage_local(
            mock_instance)
        exp_ret_val = {'filename': mock.sentinel.tmp_file}

        self.assertEqual(exp_ret_val, ret_val)
        mock_get_instance_dir.assert_called_once_with(mock_instance.name)
        mock_mkstemp.assert_called_once_with(
            dir=mock_get_instance_dir.return_value)
        mock_close.assert_called_once_with(mock.sentinel.tmp_fd)

    @mock.patch.object(os.path, 'exists')
    def test_check_instance_shared_storage_remote(self, mock_exists):
        check_data = dict(filename=mock.sentinel.filename)
        ret_val = self._pathutils.check_instance_shared_storage_remote(
            check_data)

        self.assertEqual(mock_exists.return_value, ret_val)

    @mock.patch.object(fileutils, 'delete_if_exists')
    def test_check_instance_shared_storage_cleanup(self,
                                                   mock_delete_if_exists):
        check_data = dict(filename=mock.sentinel.filename)
        self._pathutils.check_instance_shared_storage_cleanup(check_data)

        mock_delete_if_exists.assert_called_once_with(mock.sentinel.filename)

    @mock.patch.object(pathutils.PathUtils, 'get_instance_dir')
    def test_get_instance_snapshot_dir(self, mock_get_instance_dir):
        mock_get_instance_dir.return_value = self.fake_instance_dir
        response = self._pathutils.get_instance_snapshot_dir(
            self.fake_instance_name)

        expected_path = os.path.join(self.fake_instance_dir, 'Snapshots')
        self.assertEqual(expected_path, response)
        mock_get_instance_dir.assert_called_once_with(self.fake_instance_name,
                                                      create_dir=False)

    @mock.patch.object(pathutils.PathUtils, 'get_instance_dir')
    def test_get_instance_virtual_machines_dir(self, mock_get_instance_dir):
        mock_get_instance_dir.return_value = self.fake_instance_dir
        response = self._pathutils.get_instance_virtual_machines_dir(
            self.fake_instance_name)

        expected_path = os.path.join(self.fake_instance_dir,
                                     'Virtual Machines')
        self.assertEqual(expected_path, response)
        mock_get_instance_dir.assert_called_once_with(self.fake_instance_name,
                                                      create_dir=False)

    @mock.patch.object(pathutils.PathUtils, 'copy_folder_files')
    @mock.patch.object(pathutils.PathUtils,
                       'get_instance_virtual_machines_dir')
    def test_copy_vm_config_files(self, mock_get_inst_vm_dir, mock_copy_files):
        self._pathutils.copy_vm_config_files(mock.sentinel.instance_name,
                                             mock.sentinel.dest_dir)

        mock_get_inst_vm_dir.assert_called_once_with(
            mock.sentinel.instance_name)
        mock_copy_files.assert_called_once_with(
            mock_get_inst_vm_dir.return_value, mock.sentinel.dest_dir)

    @mock.patch('os.listdir')
    def test_get_vm_config_file(self, mock_listdir):
        config_file = '81027A62-7187-4EC4-AFF5-9CA853BF7C68.vmcx'
        mock_listdir.return_value = [config_file]

        response = self._pathutils.get_vm_config_file(self.fake_instance_dir)

        expected_path = os.path.join(self.fake_instance_dir, config_file)
        self.assertEqual(expected_path, response)
        mock_listdir.assert_called_once_with(self.fake_instance_dir)

    @mock.patch('os.listdir')
    def test_get_vm_config_file_exception(self, mock_listdir):
        mock_listdir.return_value = ['fake_file']

        self.assertRaises(exception.NotFound,
                          self._pathutils.get_vm_config_file,
                          mock.sentinel.instances_path)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

import nova.conf

hyperv_opts = [
    cfg.IntOpt('evacuate_task_state_timeout',
               default=600,
               help='Number of seconds to wait for an instance to be '
                    'evacuated during host maintenance.'),
    cfg.IntOpt('cluster_event_check_interval',
               deprecated_for_removal=True,
               deprecated_since="5.0.1",
               default=2),
    cfg.BoolOpt('instance_automatic_shutdown',
                default=False,
                help='Automatically shutdown instances when the host is '
                     'shutdown. By default, instances will be saved, which '
                     'adds a disk overhead. Changing this option will not '
                     'affect existing instances.'),
    cfg.IntOpt('instance_live_migration_timeout',
               default=300,
               min=0,
               help='Number of seconds to wait for an instance to be '
                    'live migrated (Only applies to clustered instances '
                    'for the moment).'),
    cfg.IntOpt('max_failover_count',
               default=1,
               min=1,
               help="The maximum number of failovers that can occur in the "
                    "failover_period timeframe per VM. Once a VM's number "
                    "failover reaches this number, the VM will simply end up "
                    "in a Failed state."),
    cfg.IntOpt('failover_period',
               default=6,
               min=1,
               help="The number of hours in which the max_failover_count "
                    "number of failovers can occur."),
    cfg.BoolOpt('recreate_ports_on_failover',
                default=True,
                help="When enabled, the ports will be recreated for failed "
                     "over instances. This ensures that we're not left with "
                     "a stale port."),
    cfg.BoolOpt('auto_failback',
                default=True,
                help="Allow the VM the failback to its original host once it "
                     "is available."),
    cfg.BoolOpt('force_destroy_instances',
                default=False,
                help="If this option is enabled, instance destroy requests "
                     "are executed immediately, regardless of instance "
                     "pending tasks. In some situations, the destroy "
                     "operation will fail (e.g. due to file locks), "
                     "requiring subsequent retries."),
    cfg.BoolOpt('move_disks_on_cold_migration',
                default=True,
                help="Move the instance files to the instance dir configured "
                     "on the destination host. You may consider disabling "
                     "this when using multiple CSVs or shares and you wish "
                     "the source location to be preserved."),
]

coordination_opts = [
    cfg.StrOpt('backend_url',
               default='file:///C:/OpenStack/Lock',
               help='The backend URL to use for distributed coordination.'),
]

CONF = nova.conf.CONF
CONF.register_opts(coordination_opts, 'coordination')
CONF.register_opts(hyperv_opts, 'hyperv')


def list_opts():
    return [('coordination', coordination_opts),
            ('hyperv', hyperv_opts)]



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from nova.compute import vm_states
from nova import exception
from nova import test as nova_test
from os_win import constants as os_win_const

from compute_hyperv.nova.cluster import livemigrationops
from compute_hyperv.nova import livemigrationops as base_livemigrationops
from compute_hyperv.tests import fake_instance
from compute_hyperv.tests.unit import test_base


@ddt.ddt
class ClusterLiveMigrationOpsTestCase(test_base.HyperVBaseTestCase):
    """Unit tests for the Hyper-V Cluster LivemigrationOps class."""

    _autospec_classes = [
        base_livemigrationops.volumeops.VolumeOps,
    ]

    def setUp(self):
        super(ClusterLiveMigrationOpsTestCase, self).setUp()
        self._fake_context = 'fake_context'
        self.livemigrops = livemigrationops.ClusterLiveMigrationOps()
        self._clustutils = self.livemigrops._clustutils

    def test_is_instance_clustered(self):
        ret = self.livemigrops.is_instance_clustered(
            mock.sentinel.instance)

        self.assertEqual(
            self.livemigrops._clustutils.vm_exists.return_value, ret)

    def test_live_migration_in_cluster(self):
        migr_timeout = 10
        self.flags(instance_live_migration_timeout=migr_timeout,
                   group='hyperv')

        mock_instance = fake_instance.fake_instance_obj(self._fake_context)
        self.livemigrops._clustutils.vm_exists.return_value = True
        post_method = mock.MagicMock()
        dest = 'fake_dest'
        node_names = [dest, 'fake_node2']
        get_nodes = self.livemigrops._clustutils.get_cluster_node_names
        get_nodes.return_value = node_names

        self.livemigrops.live_migration(
            self._fake_context, mock_instance, dest, post_method,
            mock.sentinel.recover_method,
            block_migration=mock.sentinel.block_migration,
            migrate_data=mock.sentinel.migrate_data)

        clustutils = self.livemigrops._clustutils
        clustutils.live_migrate_vm.assert_called_once_with(
            mock_instance.name, dest, migr_timeout)
        post_method.assert_called_once_with(
            self._fake_context, mock_instance, dest,
            mock.sentinel.block_migration, mock.sentinel.migrate_data)

    @mock.patch.object(livemigrationops.ClusterLiveMigrationOps,
                       '_check_failed_instance_migration')
    def test_live_migration_in_cluster_exception(self, mock_check_migr):
        mock_instance = fake_instance.fake_instance_obj(self._fake_context)
        self.livemigrops._clustutils.vm_exists.return_value = True
        recover_method = mock.MagicMock()
        dest = 'fake_dest'
        node_names = [dest, 'fake_node2']
        get_nodes = self.livemigrops._clustutils.get_cluster_node_names
        get_nodes.return_value = node_names
        clustutils = self.livemigrops._clustutils
        clustutils.live_migrate_vm.side_effect = nova_test.TestingException

        self.livemigrops.live_migration(
            self._fake_context, mock_instance, dest, mock.sentinel.post_method,
            recover_method,
            block_migration=mock.sentinel.block_migration,
            migrate_data=mock.sentinel.migrate_data)

        mock_check_migr.assert_called_once_with(
            mock_instance,
            expected_state=os_win_const.CLUSTER_GROUP_ONLINE)

        recover_method.assert_called_once_with(
            self._fake_context, mock_instance, dest,
            mock.sentinel.migrate_data)

    @mock.patch.object(base_livemigrationops.LiveMigrationOps,
                       'live_migration')
    def test_live_migration_outside_cluster(self, mock_super_live_migration):
        mock_instance = fake_instance.fake_instance_obj(self._fake_context)
        self.livemigrops._clustutils.vm_exists.return_value = True
        dest = 'fake_dest'
        node_names = ['fake_node1', 'fake_node2']
        get_nodes = self.livemigrops._clustutils.get_cluster_node_names
        get_nodes.return_value = node_names

        self.livemigrops.live_migration(
            self._fake_context, mock_instance, dest, mock.sentinel.post_method,
            mock.sentinel.recover_method, block_migration=False,
            migrate_data=None)

        mock_super_live_migration.assert_called_once_with(
            self._fake_context, mock_instance, dest, mock.sentinel.post_method,
            mock.sentinel.recover_method, False, None)

    @ddt.data({},
              {'state': os_win_const.CLUSTER_GROUP_PENDING,
               'expected_invalid_state': True},
              {'migration_queued': True,
               'expected_invalid_state': True},
              {'owner_node': 'some_other_node',
               'expected_invalid_state': True})
    @ddt.unpack
    def test_check_failed_instance_migration(
            self, state=os_win_const.CLUSTER_GROUP_ONLINE,
            owner_node='source_node', migration_queued=False,
            expected_invalid_state=False):
        state_info = dict(owner_node=owner_node.upper(),
                          state=state,
                          migration_queued=migration_queued)
        self._clustutils.get_cluster_group_state_info.return_value = (
            state_info)
        self._clustutils.get_node_name.return_value = 'source_node'

        mock_instance = mock.Mock()

        if expected_invalid_state:
            self.assertRaises(
                exception.InstanceInvalidState,
                self.livemigrops._check_failed_instance_migration,
                mock_instance,
                os_win_const.CLUSTER_GROUP_ONLINE)
            self.assertEqual(vm_states.ERROR, mock_instance.vm_state)
        else:
            self.livemigrops._check_failed_instance_migration(
                mock_instance, os_win_const.CLUSTER_GROUP_ONLINE)

        self._clustutils.get_cluster_group_state_info.assert_called_once_with(
            mock_instance.name)
        self._clustutils.get_node_name.assert_called_once_with()

    def test_pre_live_migration_clustered(self):
        self.livemigrops.pre_live_migration(self._fake_context,
                                            mock.sentinel.fake_instance,
                                            mock.sentinel.bdi,
                                            mock.sentinel.network_info)

        fake_conn_vol = self.livemigrops._volumeops.connect_volumes
        fake_conn_vol.assert_called_once_with(mock.sentinel.bdi)

    @mock.patch.object(base_livemigrationops.LiveMigrationOps,
                       'pre_live_migration')
    def test_pre_live_migration_not_clustered(self, mock_pre_live_migration):
        self.livemigrops._clustutils.vm_exists.return_value = False
        self.livemigrops.pre_live_migration(self._fake_context,
                                            mock.sentinel.fake_instance,
                                            mock.sentinel.bdi,
                                            mock.sentinel.network_info)

        mock_pre_live_migration.assert_called_once_with(
            self._fake_context, mock.sentinel.fake_instance,
            mock.sentinel.bdi, mock.sentinel.network_info)

    @mock.patch.object(base_livemigrationops.LiveMigrationOps,
                       'post_live_migration')
    def test_post_live_migration_clustered(self, mock_post_live_migration):
        self.livemigrops.post_live_migration(self._fake_context,
                                             mock.sentinel.fake_instance,
                                             mock.sentinel.bdi,
                                             mock.sentinel.migrate_data)

        self.assertFalse(mock_post_live_migration.called)

    @mock.patch.object(base_livemigrationops.LiveMigrationOps,
                       'post_live_migration')
    def test_post_live_migration_not_clustered(self, mock_post_live_migration):
        self.livemigrops._clustutils.vm_exists.return_value = False
        self.livemigrops.post_live_migration(self._fake_context,
                                             mock.sentinel.fake_instance,
                                             mock.sentinel.bdi,
                                             mock.sentinel.migrate_data)

        mock_post_live_migration.assert_called_once_with(
            self._fake_context, mock.sentinel.fake_instance,
            mock.sentinel.bdi,
            mock.sentinel.migrate_data)



# Copyright 2018 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import exception

from compute_hyperv.nova.cluster import volumeops
from compute_hyperv.nova import constants
from compute_hyperv.nova import volumeops as base_volumeops
from compute_hyperv.tests.unit import test_base


class ClusterVolumeOpsTestCase(test_base.HyperVBaseTestCase):
    _autospec_classes = [
        base_volumeops.cinder.API,
    ]

    def setUp(self):
        super(ClusterVolumeOpsTestCase, self).setUp()
        self._volumeops = volumeops.ClusterVolumeOps()

    def test_loaded_volume_drivers(self):
        self.assertEqual(set([constants.STORAGE_PROTOCOL_SMBFS]),
                         set(self._volumeops.volume_drivers.keys()))

    def test_get_blacklisted_volume_driver(self):
        conn_info = dict(driver_volume_type=constants.STORAGE_PROTOCOL_ISCSI)

        self.assertRaises(
            exception.VolumeDriverNotFound,
            self._volumeops._get_volume_driver,
            conn_info)

    def test_get_supported_volume_driver(self):
        conn_info = dict(driver_volume_type=constants.STORAGE_PROTOCOL_SMBFS)
        drv = self._volumeops._get_volume_driver(conn_info)

        self.assertIsInstance(drv, base_volumeops.SMBFSVolumeDriver)



# Copyright (c) 2010 Cloud.com, Inc
# Copyright (c) 2012 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A Hyper-V Nova Compute driver.
"""

import functools
import platform
import sys

from nova import context as nova_context
from nova import exception
from nova.image import glance
from nova import objects
from nova.virt import driver
from os_win import exceptions as os_win_exc
from os_win import utilsfactory
from oslo_log import log as logging
import six

from compute_hyperv.nova import coordination
from compute_hyperv.nova import eventhandler
from compute_hyperv.nova import hostops
from compute_hyperv.nova import imagecache
from compute_hyperv.nova import livemigrationops
from compute_hyperv.nova import migrationops
from compute_hyperv.nova import pathutils
from compute_hyperv.nova import rdpconsoleops
from compute_hyperv.nova import serialconsoleops
from compute_hyperv.nova import snapshotops
from compute_hyperv.nova import vmops
from compute_hyperv.nova import volumeops


LOG = logging.getLogger(__name__)


def convert_exceptions(function, exception_map):
    expected_exceptions = tuple(exception_map.keys())

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except expected_exceptions as ex:
            raised_exception = exception_map.get(type(ex))
            if not raised_exception:
                # exception might be a subclass of an expected exception.
                for expected in expected_exceptions:
                    if isinstance(ex, expected):
                        raised_exception = exception_map[expected]
                        break

            exc_info = sys.exc_info()
            # NOTE(claudiub): Python 3 raises the exception object given as
            # the second argument in six.reraise.
            # The original message will be maintained by passing the original
            # exception.
            exc = raised_exception(six.text_type(exc_info[1]))
            six.reraise(raised_exception, exc, exc_info[2])
    return wrapper


def decorate_all_methods(decorator, *args, **kwargs):
    def decorate(cls):
        for attr in cls.__dict__:
            class_member = getattr(cls, attr)
            if callable(class_member):
                setattr(cls, attr, decorator(class_member, *args, **kwargs))
        return cls

    return decorate


exception_conversion_map = {
    # expected_exception: converted_exception
    os_win_exc.OSWinException: exception.NovaException,
    os_win_exc.HyperVVMNotFoundException: exception.InstanceNotFound,
}

# NOTE(claudiub): the purpose of the decorator below is to prevent any
# os_win exceptions (subclasses of OSWinException) to leak outside of the
# HyperVDriver.


@decorate_all_methods(convert_exceptions, exception_conversion_map)
class HyperVDriver(driver.ComputeDriver):
    capabilities = {
        "has_imagecache": True,
        "supports_evacuate": True,
        "supports_migrate_to_same_host": False,
        "supports_attach_interface": True,
        "supports_device_tagging": True,
        "supports_tagged_attach_interface": True,
        "supports_tagged_attach_volume": True,
        "supports_extend_volume": True,
        "supports_multiattach": False,
        "supports_trusted_certs": True,

        # Supported image types
        "supports_image_type_vhd": True,
        "supports_image_type_vhdx": True,
    }

    use_coordination = False

    def __init__(self, virtapi):
        # check if the current version of Windows is supported before any
        # further driver initialisation.
        self._check_minimum_windows_version()

        # We'll initialize coordination as early as possible, avoiding
        # the risk of using locks before the mechanism is enabled.
        if self.use_coordination:
            coordination.COORDINATOR.start()

        super(HyperVDriver, self).__init__(virtapi)

        self._hostops = hostops.HostOps()
        self._volumeops = volumeops.VolumeOps()
        self._vmops = vmops.VMOps(virtapi)
        self._snapshotops = snapshotops.SnapshotOps()
        self._livemigrationops = livemigrationops.LiveMigrationOps()
        self._migrationops = migrationops.MigrationOps()
        self._rdpconsoleops = rdpconsoleops.RDPConsoleOps()
        self._serialconsoleops = serialconsoleops.SerialConsoleOps()
        self._imagecache = imagecache.ImageCache()
        self._image_api = glance.API()
        self._pathutils = pathutils.PathUtils()
        self._event_handler = eventhandler.InstanceEventHandler()

    def _check_minimum_windows_version(self):
        hostutils = utilsfactory.get_hostutils()
        if not hostutils.check_min_windows_version(6, 2):
            # the version is of Windows is older than Windows Server 2012 R2.
            # Log an error, letting users know that this version is not
            # supported any longer.
            LOG.error('You are running nova-compute on an unsupported '
                      'version of Windows (older than Windows / Hyper-V '
                      'Server 2012). The support for this version of '
                      'Windows has been removed in Mitaka.')
            raise exception.HypervisorTooOld(version='6.2')
        elif not hostutils.check_min_windows_version(6, 3):
            # TODO(claudiub): replace the warning with an exception in Rocky.
            LOG.warning('You are running nova-compute on Windows / Hyper-V '
                        'Server 2012. The support for this version of Windows '
                        'has been deprecated In Queens, and will be removed '
                        'in Rocky.')

    def init_host(self, host):
        self._serialconsoleops.start_console_handlers()

        self._set_event_handler_callbacks()
        self._event_handler.start_listener()

        instance_path = self._pathutils.get_instances_dir()
        self._pathutils.check_create_dir(instance_path)

    def _set_event_handler_callbacks(self):
        # Subclasses may override this.
        self._event_handler.add_callback(self.emit_event)
        self._event_handler.add_callback(
            self._vmops.instance_state_change_callback)

    def list_instance_uuids(self):
        return self._vmops.list_instance_uuids()

    def list_instances(self):
        return self._vmops.list_instances()

    def spawn(self, context, instance, image_meta, injected_files,
              admin_password, allocations, network_info=None,
              block_device_info=None, power_on=True, accel_info=None):
        image_meta = self._recreate_image_meta(context, instance, image_meta)
        self._vmops.spawn(context, instance, image_meta, injected_files,
                          admin_password, network_info, block_device_info,
                          power_on)

    def reboot(self, context, instance, network_info, reboot_type,
               block_device_info=None, bad_volumes_callback=None,
               accel_info=None):
        self._vmops.reboot(instance, network_info, reboot_type)

    def destroy(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, destroy_secrets=True):
        self._vmops.destroy(instance, network_info, block_device_info,
                            destroy_disks)

    def cleanup(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, migrate_data=None, destroy_vifs=True,
                destroy_secrets=True):
        """Cleanup after instance being destroyed by Hypervisor."""
        self.unplug_vifs(instance, network_info)

    def get_info(self, instance, use_cache=True):
        return self._vmops.get_info(instance)

    def attach_volume(self, context, connection_info, instance, mountpoint,
                      disk_bus=None, device_type=None, encryption=None):
        self._volumeops.attach_volume(context,
                                      connection_info,
                                      instance,
                                      update_device_metadata=True)

    def detach_volume(self, context, connection_info, instance, mountpoint,
                      encryption=None):
        context = nova_context.get_admin_context()
        # The nova compute manager only updates the device metadata in
        # case of tagged devices. We're including untagged devices as well.
        self._volumeops.detach_volume(context,
                                      connection_info,
                                      instance,
                                      update_device_metadata=True)

    def extend_volume(self, connection_info, instance, requested_size):
        self._volumeops.extend_volume(connection_info)

    def get_volume_connector(self, instance):
        return self._volumeops.get_volume_connector()

    def get_available_resource(self, nodename):
        return self._hostops.get_available_resource()

    def get_available_nodes(self, refresh=False):
        return [platform.node()]

    def host_power_action(self, action):
        return self._hostops.host_power_action(action)

    def snapshot(self, context, instance, image_id, update_task_state):
        self._snapshotops.snapshot(context, instance, image_id,
                                   update_task_state)

    def volume_snapshot_create(self, context, instance, volume_id,
                               create_info):
        self._volumeops.volume_snapshot_create(context, instance, volume_id,
                                               create_info)

    def volume_snapshot_delete(self, context, instance, volume_id,
                               snapshot_id, delete_info):
        self._volumeops.volume_snapshot_delete(context, instance, volume_id,
                                               snapshot_id, delete_info)

    def pause(self, instance):
        self._vmops.pause(instance)

    def unpause(self, instance):
        self._vmops.unpause(instance)

    def suspend(self, context, instance):
        self._vmops.suspend(instance)

    def resume(self, context, instance, network_info, block_device_info=None):
        self._vmops.resume(instance)

    def power_off(self, instance, timeout=0, retry_interval=0):
        self._vmops.power_off(instance, timeout, retry_interval)

    def power_on(self, context, instance, network_info,
                 block_device_info=None, accel_info=None):
        self._vmops.power_on(instance, block_device_info, network_info)

    def resume_state_on_host_boot(self, context, instance, network_info,
                                  block_device_info=None):
        """Resume guest state when a host is booted."""
        self._vmops.resume_state_on_host_boot(context, instance, network_info,
                                              block_device_info)

    def live_migration(self, context, instance, dest, post_method,
                       recover_method, block_migration=False,
                       migrate_data=None):
        self._livemigrationops.live_migration(context, instance, dest,
                                              post_method, recover_method,
                                              block_migration, migrate_data)

    def rollback_live_migration_at_destination(self, context, instance,
                                               network_info,
                                               block_device_info,
                                               destroy_disks=True,
                                               migrate_data=None):
        self.destroy(context, instance, network_info, block_device_info,
                     destroy_disks=destroy_disks)

    def pre_live_migration(self, context, instance, block_device_info,
                           network_info, disk_info, migrate_data):
        self._livemigrationops.pre_live_migration(context, instance,
                                                  block_device_info,
                                                  network_info)
        return migrate_data

    def post_live_migration(self, context, instance, block_device_info,
                            migrate_data=None):
        self._livemigrationops.post_live_migration(context, instance,
                                                   block_device_info,
                                                   migrate_data)

    def post_live_migration_at_source(self, context, instance, network_info):
        """Unplug VIFs from networks at source."""
        self._vmops.unplug_vifs(instance, network_info)

    def post_live_migration_at_destination(self, context, instance,
                                           network_info,
                                           block_migration=False,
                                           block_device_info=None):
        self._livemigrationops.post_live_migration_at_destination(
            context,
            instance,
            network_info,
            block_migration)

    def check_can_live_migrate_destination(self, context, instance,
                                           src_compute_info, dst_compute_info,
                                           block_migration=False,
                                           disk_over_commit=False):
        return self._livemigrationops.check_can_live_migrate_destination(
            context, instance, src_compute_info, dst_compute_info,
            block_migration, disk_over_commit)

    def cleanup_live_migration_destination_check(self, context,
                                                 dest_check_data):
        self._livemigrationops.cleanup_live_migration_destination_check(
            context, dest_check_data)

    def check_can_live_migrate_source(self, context, instance,
                                      dest_check_data, block_device_info=None):
        return self._livemigrationops.check_can_live_migrate_source(
            context, instance, dest_check_data)

    def get_instance_disk_info(self, instance, block_device_info=None):
        pass

    def plug_vifs(self, instance, network_info):
        """Plug VIFs into networks."""
        self._vmops.plug_vifs(instance, network_info)

    def unplug_vifs(self, instance, network_info):
        """Unplug VIFs from networks."""
        self._vmops.unplug_vifs(instance, network_info)

    def ensure_filtering_rules_for_instance(self, instance, network_info):
        LOG.debug("ensure_filtering_rules_for_instance called",
                  instance=instance)

    def unfilter_instance(self, instance, network_info):
        LOG.debug("unfilter_instance called", instance=instance)

    def migrate_disk_and_power_off(self, context, instance, dest,
                                   flavor, network_info,
                                   block_device_info=None,
                                   timeout=0, retry_interval=0):
        return self._migrationops.migrate_disk_and_power_off(context,
                                                             instance, dest,
                                                             flavor,
                                                             network_info,
                                                             block_device_info,
                                                             timeout,
                                                             retry_interval)

    def confirm_migration(self, context, migration, instance, network_info):
        self._migrationops.confirm_migration(context, migration,
                                             instance, network_info)

    def finish_revert_migration(self, context, instance, network_info,
                                migration, block_device_info=None,
                                power_on=True):
        self._migrationops.finish_revert_migration(context, instance,
                                                   network_info,
                                                   block_device_info, power_on)

    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance,
                         allocations, block_device_info=None, power_on=True):
        image_meta = self._recreate_image_meta(context, instance, image_meta)
        self._migrationops.finish_migration(context, migration, instance,
                                            disk_info, network_info,
                                            image_meta, resize_instance,
                                            block_device_info, power_on)

    def get_host_ip_addr(self):
        return self._hostops.get_host_ip_addr()

    def get_host_uptime(self):
        return self._hostops.get_host_uptime()

    def get_rdp_console(self, context, instance):
        return self._rdpconsoleops.get_rdp_console(instance)

    def get_serial_console(self, context, instance):
        return self._serialconsoleops.get_serial_console(instance.name)

    def get_console_output(self, context, instance):
        return self._serialconsoleops.get_console_output(instance.name)

    def manage_image_cache(self, context, all_instances):
        self._imagecache.update(context, all_instances)

    def cache_image(self, context, image_id):
        image_path, fetched = self._imagecache.cache_image(
            context, image_id)
        return fetched

    def attach_interface(self, context, instance, image_meta, vif):
        self._vmops.attach_interface(context, instance, vif)

    def detach_interface(self, context, instance, vif):
        # The device metadata gets updated outside the driver.
        return self._vmops.detach_interface(instance, vif)

    def rescue(self, context, instance, network_info, image_meta,
               rescue_password, block_device_info):
        image_meta = self._recreate_image_meta(context, instance, image_meta)
        self._vmops.rescue_instance(context, instance, network_info,
                                    image_meta, rescue_password)

    def unrescue(
        self,
        context: nova_context.RequestContext,
        instance: 'objects.Instance',
    ):
        self._vmops.unrescue_instance(instance)

    def host_maintenance_mode(self, host, mode):
        return self._hostops.host_maintenance_mode(host, mode)

    def _recreate_image_meta(self, context, instance, image_meta):
        # TODO(claudiub): Cleanup this method. instance.system_metadata might
        # already contain all the image metadata properties we need anyways.
        if image_meta.obj_attr_is_set("id"):
            image_ref = image_meta.id
        else:
            image_ref = instance.system_metadata['image_base_image_ref']

        if image_ref:
            image_meta = self._image_api.get(context, image_ref)
        else:
            # boot from volume does not have an image_ref.
            image_meta = image_meta.obj_to_primitive()['nova_object.data']
            image_meta['properties'] = {k.replace('image_', '', 1): v for k, v
                                        in instance.system_metadata.items()}
        image_meta["id"] = image_ref
        return image_meta

    def check_instance_shared_storage_local(self, context, instance):
        """Check if instance files located on shared storage.

        This runs check on the destination host, and then calls
        back to the source host to check the results.

        :param context: security context
        :param instance: nova.objects.instance.Instance object
        :returns: A dict containing the tempfile info.
        """
        return self._pathutils.check_instance_shared_storage_local(instance)

    def check_instance_shared_storage_remote(self, context, data):
        return self._pathutils.check_instance_shared_storage_remote(data)

    def check_instance_shared_storage_cleanup(self, context, data):
        return self._pathutils.check_instance_shared_storage_cleanup(data)

    def update_provider_tree(self, provider_tree, nodename, allocations=None):
        inventory = provider_tree.data(nodename).inventory
        alloc_ratios = self._get_allocation_ratios(inventory)

        self._hostops.update_provider_tree(
            provider_tree, nodename, alloc_ratios, allocations)



# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile
import time

from nova import exception
from os_win import exceptions as os_win_exc
from os_win.utils import pathutils
from os_win import utilsfactory
from oslo_log import log as logging
from oslo_utils import fileutils
from oslo_utils import uuidutils

from compute_hyperv.i18n import _
import compute_hyperv.nova.conf
from compute_hyperv.nova import constants

LOG = logging.getLogger(__name__)

CONF = compute_hyperv.nova.conf.CONF

ERROR_INVALID_NAME = 123

# NOTE(claudiub): part of the pre-existing PathUtils is nova-specific and
# it does not belong in the os-win library. In order to ensure the same
# functionality with the least amount of changes necessary, adding as a mixin
# the os_win.pathutils.PathUtils class into this PathUtils.


class PathUtils(pathutils.PathUtils):

    _CSV_FOLDER = 'ClusterStorage\\'

    def __init__(self):
        super(PathUtils, self).__init__()
        self._vmutils = utilsfactory.get_vmutils()

    def copy_folder_files(self, src_dir, dest_dir):
        """Copies the files of the given src_dir to dest_dir.

        It will ignore any nested folders.

        :param src_dir: Given folder from which to copy files.
        :param dest_dir: Folder to which to copy files.
        """

        # NOTE(claudiub): this will have to be moved to os-win.

        for fname in os.listdir(src_dir):
            src = os.path.join(src_dir, fname)
            # ignore subdirs.
            if os.path.isfile(src):
                self.copy(src, os.path.join(dest_dir, fname))

    def get_instances_dir(self, remote_server=None):
        local_instance_path = os.path.normpath(CONF.instances_path)

        if remote_server and not local_instance_path.startswith(r'\\'):
            if CONF.hyperv.instances_path_share:
                path = CONF.hyperv.instances_path_share
            else:
                # In this case, we expect the instance dir to have the same
                # location on the remote server.
                path = local_instance_path
            return self.get_remote_path(remote_server, path)
        else:
            return local_instance_path

    def get_remote_path(self, remote_server, remote_path):
        if remote_path.startswith('\\\\'):
            return remote_path

        # Use an administrative share
        remote_unc_path = ('\\\\%(remote_server)s\\%(path)s' %
                           dict(remote_server=remote_server,
                                path=remote_path.replace(':', '$')))

        csv_location = '\\'.join([os.getenv('SYSTEMDRIVE', 'C:'),
                                  self._CSV_FOLDER])
        if remote_path.lower().startswith(csv_location.lower()):
            # the given remote_path is a CSV path.
            # Return remote_path as the local path.
            LOG.debug("Remote path %s is on a CSV. Returning as a local path.",
                      remote_path)
            return remote_path

        LOG.debug('Returning UNC path %(unc_path)s for host %(host)s.',
                  dict(unc_path=remote_unc_path, host=remote_server))
        return remote_unc_path

    def _get_instances_sub_dir(self, dir_name, remote_server=None,
                               create_dir=True, remove_dir=False):
        instances_path = self.get_instances_dir(remote_server)
        path = os.path.join(instances_path, dir_name)
        self.check_dir(path, create_dir=create_dir, remove_dir=remove_dir)

        return path

    def check_dir(self, path, create_dir=False, remove_dir=False):
        try:
            if remove_dir:
                self.check_remove_dir(path)
            if create_dir:
                self.check_create_dir(path)
        except WindowsError as ex:
            if ex.winerror == ERROR_INVALID_NAME:
                raise exception.AdminRequired(_(
                    "Cannot access \"%(path)s\", make sure the "
                    "path exists and that you have the proper permissions. "
                    "In particular Nova-Compute must not be executed with the "
                    "builtin SYSTEM account or other accounts unable to "
                    "authenticate on a remote host.") % {'path': path})
            raise

    def get_instance_migr_revert_dir(self, instance_path, create_dir=False,
                                     remove_dir=False):
        dir_name = '%s_revert' % instance_path
        self.check_dir(dir_name, create_dir, remove_dir)
        return dir_name

    def get_instance_dir(self, instance_name, remote_server=None,
                         create_dir=True, remove_dir=False):
        instance_dir = self._get_instances_sub_dir(
            instance_name, remote_server,
            create_dir=False, remove_dir=False)

        # In some situations, the instance files may reside at a different
        # location than the configured one.
        if not os.path.exists(instance_dir):
            vmutils = (self._vmutils if not remote_server
                       else utilsfactory.get_vmutils(remote_server))
            try:
                instance_dir = vmutils.get_vm_config_root_dir(
                    instance_name)
                if remote_server:
                    instance_dir = self.get_remote_path(remote_server,
                                                        instance_dir)
                LOG.info("Found instance dir at non-default location: %s",
                         instance_dir)
            except os_win_exc.HyperVVMNotFoundException:
                pass

        self.check_dir(instance_dir,
                       create_dir=create_dir,
                       remove_dir=remove_dir)
        return instance_dir

    def _lookup_vhd_path(self, instance_name, vhd_path_func,
                         *args, **kwargs):
        vhd_path = None
        for format_ext in ['vhd', 'vhdx']:
            test_path = vhd_path_func(instance_name, format_ext,
                                      *args, **kwargs)
            if self.exists(test_path):
                vhd_path = test_path
                break
        return vhd_path

    def lookup_root_vhd_path(self, instance_name, rescue=False):
        return self._lookup_vhd_path(instance_name, self.get_root_vhd_path,
                                     rescue)

    def lookup_configdrive_path(self, instance_name, rescue=False):
        configdrive_path = None
        for format_ext in constants.DISK_FORMAT_MAP:
            test_path = self.get_configdrive_path(instance_name, format_ext,
                                                  rescue=rescue)
            if self.exists(test_path):
                configdrive_path = test_path
                break
        return configdrive_path

    def lookup_ephemeral_vhd_path(self, instance_name, eph_name):
        return self._lookup_vhd_path(instance_name,
                                     self.get_ephemeral_vhd_path,
                                     eph_name)

    def get_root_vhd_path(self, instance_name, format_ext=None, rescue=False):
        instance_path = self.get_instance_dir(instance_name)
        image_name = 'root'
        if rescue:
            image_name += '-rescue'
        if format_ext:
            image_name += '.' + format_ext.lower()
        return os.path.join(instance_path, image_name)

    def get_configdrive_path(self, instance_name, format_ext,
                             remote_server=None, rescue=False):
        instance_path = self.get_instance_dir(instance_name, remote_server)
        configdrive_image_name = 'configdrive'
        if rescue:
            configdrive_image_name += '-rescue'
        return os.path.join(instance_path,
                            configdrive_image_name + '.' + format_ext.lower())

    def get_ephemeral_vhd_path(self, instance_name, format_ext, eph_name):
        instance_path = self.get_instance_dir(instance_name)
        return os.path.join(instance_path, eph_name + '.' + format_ext.lower())

    def get_base_vhd_dir(self):
        return self._get_instances_sub_dir('_base')

    def get_export_dir(self, instance_name=None, instance_dir=None,
                       create_dir=False, remove_dir=False):
        if not instance_dir:
            instance_dir = self.get_instance_dir(instance_name,
                                                 create_dir=create_dir)

        export_dir = os.path.join(instance_dir, 'export')
        self.check_dir(export_dir, create_dir=create_dir,
                       remove_dir=remove_dir)
        return export_dir

    def get_vm_console_log_paths(self, instance_name, remote_server=None):
        instance_dir = self.get_instance_dir(instance_name,
                                             remote_server)
        console_log_path = os.path.join(instance_dir, 'console.log')
        return console_log_path, console_log_path + '.1'

    def copy_vm_console_logs(self, instance_name, dest_host):
        local_log_paths = self.get_vm_console_log_paths(
            instance_name)
        remote_log_paths = self.get_vm_console_log_paths(
            instance_name, remote_server=dest_host)

        for local_log_path, remote_log_path in zip(local_log_paths,
                                                   remote_log_paths):
            if self.exists(local_log_path):
                self.copy(local_log_path, remote_log_path)

    def get_image_path(self, image_name):
        # Note: it is possible that the path doesn't exist
        base_dir = self.get_base_vhd_dir()
        for ext in ['vhd', 'vhdx', 'iso']:
            file_path = os.path.join(base_dir,
                                     image_name + '.' + ext.lower())
            if self.exists(file_path):
                return file_path
        return None

    def get_age_of_file(self, file_name):
        return time.time() - os.path.getmtime(file_name)

    def check_dirs_shared_storage(self, src_dir, dest_dir):
        # Check if shared storage is being used by creating a temporary
        # file at the destination path and checking if it exists at the
        # source path.
        LOG.debug("Checking if %(src_dir)s and %(dest_dir)s point "
                  "to the same location.",
                  dict(src_dir=src_dir, dest_dir=dest_dir))
        with tempfile.NamedTemporaryFile(dir=dest_dir) as tmp_file:
            src_path = os.path.join(src_dir,
                                    os.path.basename(tmp_file.name))

            shared_storage = os.path.exists(src_path)
        return shared_storage

    def check_remote_instances_dir_shared(self, dest):
        # Checks if the instances dir from a remote host points
        # to the same storage location as the local instances dir.
        local_inst_dir = self.get_instances_dir()
        remote_inst_dir = self.get_instances_dir(dest)
        return self.check_dirs_shared_storage(local_inst_dir,
                                              remote_inst_dir)

    def check_instance_shared_storage_local(self, instance):
        instance_dir = self.get_instance_dir(instance.name)

        fd, tmp_file = tempfile.mkstemp(dir=instance_dir)
        LOG.debug("Creating tmpfile %s to verify with other "
                  "compute node that the instance is on "
                  "the same shared storage.",
                  tmp_file, instance=instance)
        os.close(fd)
        # We're sticking with the same dict key as the libvirt driver.
        # At some point, this may become a versioned object.
        return {"filename": tmp_file}

    def check_instance_shared_storage_remote(self, data):
        return os.path.exists(data['filename'])

    def check_instance_shared_storage_cleanup(self, data):
        fileutils.delete_if_exists(data["filename"])

    def get_instance_snapshot_dir(self, instance_name=None, instance_dir=None):
        if instance_name:
            instance_dir = self.get_instance_dir(instance_name,
                                                 create_dir=False)
        return os.path.join(instance_dir, 'Snapshots')

    def get_instance_virtual_machines_dir(self, instance_name=None,
                                          instance_dir=None):
        if instance_name:
            instance_dir = self.get_instance_dir(instance_name,
                                                 create_dir=False)
        return os.path.join(instance_dir, "Virtual Machines")

    def copy_vm_config_files(self, instance_name, dest_dir):
        """Copies the VM configuration files to the given destination folder.

        :param instance_name: the given instance's name.
        :param dest_dir: the location where the VM configuration files are
            copied to.
        """
        src_dir = self.get_instance_virtual_machines_dir(instance_name)
        self.copy_folder_files(src_dir, dest_dir)

    def get_vm_config_file(self, path):
        for dir_file in os.listdir(path):
            file_name, file_ext = os.path.splitext(dir_file)
            if (file_ext.lower() in ['.vmcx', '.xml'] and
                    uuidutils.is_uuid_like(file_name)):

                config_file = os.path.join(path, dir_file)
                LOG.debug("Found VM config file: %s", config_file)
                return config_file

        raise exception.NotFound(
            _("Folder %s does not contain any VM config data file.") % path)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for our unit tests.

Allows overriding of flags for use of fakes, and some black magic for
inline callbacks.

"""

import eventlet
eventlet.monkey_patch(os=False)
import inspect
from unittest import mock

import fixtures
from nova.tests import fixtures as nova_fixtures
from oslo_log.fixture import logging_error as log_fixture
from oslo_log import log as logging
from oslotest import base
from oslotest import mock_fixture
import six

import compute_hyperv.nova.conf

CONF = compute_hyperv.nova.conf.CONF

logging.register_options(CONF)
CONF.set_override('use_stderr', False)

_TRUE_VALUES = ('True', 'true', '1', 'yes')


def _patch_mock_to_raise_for_invalid_assert_calls():
    def raise_for_invalid_assert_calls(wrapped):
        def wrapper(_self, name):
            valid_asserts = [
                'assert_called_with',
                'assert_called_once_with',
                'assert_has_calls',
                'assert_any_calls']

            if name.startswith('assert') and name not in valid_asserts:
                raise AttributeError('%s is not a valid mock assert method'
                                     % name)

            return wrapped(_self, name)
        return wrapper
    mock.Mock.__getattr__ = raise_for_invalid_assert_calls(
        mock.Mock.__getattr__)


# NOTE(gibi): needs to be called only once at import time
# to patch the mock lib
_patch_mock_to_raise_for_invalid_assert_calls()

# NOTE(claudiub): this needs to be called before any mock.patch calls are
# being done, and especially before any other test classes load. This fixes
# the mock.patch autospec issue:
# https://github.com/testing-cabal/mock/issues/396
mock_fixture.patch_mock_module()


class NoDBTestCase(base.BaseTestCase):
    """Test case base class for all unit tests.

    Due to the slowness of DB access, please consider deriving from
    `NoDBTestCase` first.
    """

    TIMEOUT_SCALING_FACTOR = 1
    MOCK_TOOZ = True

    def setUp(self):
        """Run before each test method to initialize test environment."""
        # Ensure BaseTestCase's ConfigureLogging fixture is disabled since
        # we're using the one from Nova (StandardLogging).
        with fixtures.EnvironmentVariable('OS_LOG_CAPTURE', '0'):
            super(NoDBTestCase, self).setUp()

        self.useFixture(mock_fixture.MockAutospecFixture())

        self.useFixture(log_fixture.get_logging_handle_error_fixture())

        self.useFixture(nova_fixtures.StandardLogging())
        self.useFixture(nova_fixtures.ConfFixture(CONF))

        # NOTE(blk-u): WarningsFixture must be after the Database fixture
        # because sqlalchemy-migrate messes with the warnings filters.
        self.useFixture(nova_fixtures.WarningsFixture())

        self.addCleanup(self._clear_attrs)
        self.policy = self.useFixture(nova_fixtures.PolicyFixture())

        self.useFixture(nova_fixtures.PoisonFunctions())

        if self.MOCK_TOOZ:
            self.patch('compute_hyperv.nova.coordination.Coordinator.start')
            self.patch('compute_hyperv.nova.coordination.Coordinator.stop')
            self.patch('compute_hyperv.nova.coordination.Coordinator.get_lock')

    def _clear_attrs(self):
        # Delete attributes that don't start with _ so they don't pin
        # memory around unnecessarily for the duration of the test
        # suite
        for key in [k for k in six.iterkeys(self.__dict__) if k[0] != '_']:
            del self.__dict__[key]

    def flags(self, **kw):
        """Override flag variables for a test."""
        group = kw.pop('group', None)
        for k, v in six.iteritems(kw):
            CONF.set_override(k, v, group)

    def patch(self, path, *args, **kwargs):
        patcher = mock.patch(path, *args, **kwargs)
        result = patcher.start()
        return result

    def assertPublicAPISignatures(self, baseinst, inst):
        def get_public_apis(inst):
            methods = {}
            for (name, value) in inspect.getmembers(inst, inspect.ismethod):
                if name.startswith("_"):
                    continue
                methods[name] = value
            return methods

        baseclass = baseinst.__class__.__name__
        basemethods = get_public_apis(baseinst)
        implmethods = get_public_apis(inst)

        extranames = []
        for name in sorted(implmethods.keys()):
            if name not in basemethods:
                extranames.append(name)

        self.assertEqual([], extranames,
                         "public APIs not listed in base class %s" %
                         baseclass)

        for name in sorted(implmethods.keys()):
            baseargs = inspect.getargspec(basemethods[name])
            implargs = inspect.getargspec(implmethods[name])

            self.assertEqual(baseargs, implargs,
                             "%s args don't match base class %s" %
                             (name, baseclass))



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova import utils
from nova.virt import event as virtevent
from os_win import constants
from os_win import utilsfactory
from oslo_log import log as logging

import compute_hyperv.nova.conf
from compute_hyperv.nova import vmops

LOG = logging.getLogger(__name__)

CONF = compute_hyperv.nova.conf.CONF


class HyperVLifecycleEvent(virtevent.LifecycleEvent):
    def __init__(self, uuid, name, transition, timestamp=None):
        super(HyperVLifecycleEvent, self).__init__(uuid, transition, timestamp)

        self.name = name


class InstanceEventHandler(object):
    _TRANSITION_MAP = {
        constants.HYPERV_VM_STATE_ENABLED: virtevent.EVENT_LIFECYCLE_STARTED,
        constants.HYPERV_VM_STATE_DISABLED: virtevent.EVENT_LIFECYCLE_STOPPED,
        constants.HYPERV_VM_STATE_PAUSED: virtevent.EVENT_LIFECYCLE_PAUSED,
        constants.HYPERV_VM_STATE_SUSPENDED:
            virtevent.EVENT_LIFECYCLE_SUSPENDED
    }

    def __init__(self):
        self._vmutils = utilsfactory.get_vmutils()
        self._listener = self._vmutils.get_vm_power_state_change_listener(
            timeframe=CONF.hyperv.power_state_check_timeframe,
            event_timeout=CONF.hyperv.power_state_event_polling_interval,
            filtered_states=list(self._TRANSITION_MAP.keys()),
            get_handler=True)

        self._vmops = vmops.VMOps()

        self._callbacks = []

    def add_callback(self, callback):
        self._callbacks.append(callback)

    def start_listener(self):
        utils.spawn_n(self._listener, self._handle_event)

    def _handle_event(self, instance_name, instance_power_state):
        # Instance uuid set by Nova. If this is missing, we assume that
        # the instance was not created by Nova and ignore the event.
        instance_uuid = self._vmops.get_instance_uuid(instance_name)
        if instance_uuid:
            self._emit_event(instance_name,
                             instance_uuid,
                             instance_power_state)
        else:
            LOG.debug("Instance uuid could not be retrieved for instance "
                      "%(instance_name)s. Instance state change event will "
                      "be ignored. Current power state: %(power_state)s.",
                      dict(instance_name=instance_name,
                           power_state=instance_power_state))

    def _emit_event(self, instance_name, instance_uuid, instance_state):
        virt_event = self._get_virt_event(instance_uuid,
                                          instance_name,
                                          instance_state)

        for callback in self._callbacks:
            utils.spawn_n(callback, virt_event)

    def _get_virt_event(self, instance_uuid, instance_name, instance_state):
        transition = self._TRANSITION_MAP[instance_state]
        return HyperVLifecycleEvent(
            uuid=instance_uuid,
            name=instance_name,
            transition=transition)



# Copyright 2014 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from os_win import utilsfactory
from oslo_utils import importutils

from compute_hyperv.tests import test


class HyperVBaseTestCase(test.NoDBTestCase):
    _autospec_classes = []

    def setUp(self):
        super(HyperVBaseTestCase, self).setUp()

        utilsfactory_patcher = mock.patch.object(
            utilsfactory, '_get_class', HyperVBaseTestCase._mock_get_class)
        utilsfactory_patcher.start()
        self.addCleanup(mock.patch.stopall)

        self._patch_autospec_classes()

    @staticmethod
    def _mock_get_class(class_type, *args, **kwargs):
        existing_classes = utilsfactory.utils_map[class_type]
        class_info = existing_classes[0]
        imported_class = importutils.import_class(class_info['path'])

        return mock.Mock(autospec=imported_class)

    def _patch_autospec_classes(self):
        for class_type in self._autospec_classes:
            mocked_class = mock.MagicMock(autospec=class_type)
            patcher = mock.patch(
                '.'.join([class_type.__module__, class_type.__name__]),
                mocked_class)
            patcher.start()
            self.addCleanup(patcher.stop)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import vm_states
from nova import exception
from nova.network import neutron
from nova import objects
from nova.virt import event as virtevent
from os_win import constants as os_win_const
from os_win import exceptions as os_win_exc

from compute_hyperv.nova.cluster import clusterops
import compute_hyperv.nova.conf
from compute_hyperv.tests import fake_instance
from compute_hyperv.tests.unit import test_base

CONF = compute_hyperv.nova.conf.CONF


@ddt.ddt
class ClusterOpsTestCase(test_base.HyperVBaseTestCase):
    """Unit tests for the Hyper-V ClusterOps class."""

    _autospec_classes = [
        clusterops.hostops.HostOps,
        neutron.API,
        clusterops.vmops.VMOps,
        clusterops.serialconsoleops.SerialConsoleOps,
        clusterops.placement_utils.PlacementUtils,
    ]

    _FAKE_INSTANCE_NAME = 'fake_instance_name'

    def setUp(self):
        super(ClusterOpsTestCase, self).setUp()
        self.context = 'fake_context'

        self.clusterops = clusterops.ClusterOps()
        self.clusterops._context = self.context

        self._clustutils = self.clusterops._clustutils
        self._network_api = self.clusterops._network_api
        self._placement = self.clusterops._placement

    def test_get_instance_host(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops.get_instance_host(mock_instance)

        self.clusterops._clustutils.get_vm_host.assert_called_once_with(
            mock_instance.name)

    def test_add_to_cluster(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops.add_to_cluster(mock_instance)

        mock_add_vm = self.clusterops._clustutils.add_vm_to_cluster
        mock_add_vm.assert_called_once_with(
            mock_instance.name, CONF.hyperv.max_failover_count,
            CONF.hyperv.failover_period, CONF.hyperv.auto_failback)
        self.assertEqual(mock_instance.uuid,
                         self.clusterops._instance_map[mock_instance.name])

    @mock.patch.object(clusterops, 'LOG')
    def test_add_to_cluster_exception(self, mock_LOG):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_add_vm = self.clusterops._clustutils.add_vm_to_cluster
        mock_add_vm.side_effect = os_win_exc.HyperVClusterException

        self.clusterops.add_to_cluster(mock_instance)
        self.assertTrue(mock_LOG.exception.called)

    def test_remove_from_cluster(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops.remove_from_cluster(mock_instance)

        self.clusterops._clustutils.vm_exists.assert_called_once_with(
            mock_instance.name)
        self.clusterops._clustutils.delete.assert_called_once_with(
            mock_instance.name)
        self.assertIsNone(self.clusterops._instance_map.get(
            mock_instance.name))

    @mock.patch.object(clusterops, 'LOG')
    def test_remove_from_cluster_exception(self, mock_LOG):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_delete = self.clusterops._clustutils.delete
        mock_delete.side_effect = os_win_exc.HyperVClusterException

        self.clusterops.remove_from_cluster(mock_instance)
        self.assertTrue(mock_LOG.exception.called)

    def test_post_migration(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops.post_migration(mock_instance)

        self.assertEqual(
            self.clusterops._instance_map[mock_instance.name],
            mock_instance.uuid)

    @mock.patch('nova.utils.spawn_n')
    def test_start_failover_listener_daemon(self, mock_spawn):
        self.clusterops.start_failover_listener_daemon()

        spawn_args = mock_spawn.call_args_list[0][0]
        self.assertEqual(
            self._clustutils.get_vm_owner_change_listener_v2.return_value,
            spawn_args[0])

        cbk = spawn_args[1]
        cbk()

        mock_spawn.assert_called_with(self.clusterops._failover_migrate)

    @mock.patch('nova.utils.spawn_n')
    @mock.patch.object(clusterops.ClusterOps, '_failover_migrate')
    @mock.patch.object(clusterops.ClusterOps, '_get_nova_instances')
    def test_reclaim_failovered_instances(self, mock_get_instances,
                                          mock_failover_migrate,
                                          mock_spawn):
        self.clusterops._this_node = 'fake_node'
        mock_instance1 = mock.MagicMock(host='other_host')
        mock_instance2 = mock.MagicMock(host=self.clusterops._this_node)
        mock_get_instances.return_value = [mock_instance1, mock_instance2]

        self.clusterops.reclaim_failovered_instances()

        self.clusterops._vmops.list_instance_uuids.assert_called_once_with()
        mock_get_instances.assert_called_once_with(
            ['id', 'uuid', 'name', 'host'],
            self.clusterops._vmops.list_instance_uuids.return_value)
        mock_spawn.assert_called_once_with(
            mock_failover_migrate,
            mock_instance1.name,
            self.clusterops._this_node)

    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops, 'LOG')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_migrate_no_instance(self, mock_get_instance_by_name,
                                          mock_LOG,
                                          mock_wait_pending_instance):
        mock_get_instance_by_name.return_value = None

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          mock.sentinel.new_host)

        mock_LOG.debug.assert_called_once_with(
            'Instance %s does not exist in nova. Skipping.',
            mock.sentinel.instance_name)
        self.assertFalse(
            self.clusterops._network_api.get_instance_nw_info.called)

    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops, 'LOG')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_migrate_migrating(self, mock_get_instance_by_name,
                                        mock_LOG, mock_wait_pending_instance):
        instance = mock_get_instance_by_name.return_value
        instance.task_state = task_states.MIGRATING

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          'new_host')

        mock_LOG.debug.assert_called_once_with(
            'Instance %s is being migrated by Nova. This '
            'will not be treated as a failover.',
            mock.sentinel.instance_name)

    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_migrate_at_source_node(self, mock_get_instance_by_name,
                                             mock_wait_pending_instance):
        instance = mock_get_instance_by_name.return_value
        instance.host = 'old_host'
        self.clusterops._this_node = instance.host

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          'new_host')

        self.clusterops._vmops.unplug_vifs.assert_called_once_with(instance,
            self.clusterops._network_api.get_instance_nw_info.return_value)

    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops, 'LOG')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_migrate_not_this_node(self, mock_get_instance_by_name,
                                            mock_LOG,
                                            mock_wait_pending_instance):
        self.clusterops._this_node = 'new_host'

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          'host')

        mock_LOG.debug.assert_called_once_with(
            'Instance %s did not failover to this node.',
            mock.sentinel.instance_name)

    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops.ClusterOps, '_failover_migrate_networks')
    @mock.patch.object(clusterops.ClusterOps, '_nova_failover_server')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_migrate_changed_host(self, mock_get_instance_by_name,
                                           mock_nova_failover_server,
                                           mock_failover_migrate_networks,
                                           mock_wait_pending_instance):
        instance = mock_get_instance_by_name.return_value
        old_host = 'old_host'
        new_host = 'new_host'
        instance.host = old_host
        self.clusterops._this_node = new_host
        self._clustutils.get_vm_host.return_value = new_host
        # Placement exceptions shouldn't break the rest of the failover logic.
        self._placement.move_compute_node_allocations.side_effect = (
            exception.NovaException)

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          new_host)

        mock_wait_pending_instance.assert_called_once_with(
            mock.sentinel.instance_name)
        self._clustutils.get_vm_host.assert_called_once_with(
            mock.sentinel.instance_name)
        mock_get_instance_by_name.assert_called_once_with(
            mock.sentinel.instance_name)
        get_inst_nw_info = self.clusterops._network_api.get_instance_nw_info
        get_inst_nw_info.assert_called_once_with(self.clusterops._context,
                                                 instance)
        mock_nova_failover_server.assert_called_once_with(instance, new_host)
        mock_failover_migrate_networks.assert_called_once_with(
            instance, old_host)
        self._placement.move_compute_node_allocations.assert_called_once_with(
            self.clusterops._context, instance, old_host, new_host,
            merge_existing=False)
        self.clusterops._vmops.plug_vifs.assert_called_once_with(
            instance, get_inst_nw_info.return_value)
        c_handler = self.clusterops._serial_console_ops.start_console_handler
        c_handler.assert_called_once_with(mock.sentinel.instance_name)

    @ddt.data({},
              {'recreate_ports_on_failover': True})
    @ddt.unpack
    @mock.patch.object(clusterops.ClusterOps, '_wait_for_pending_instance')
    @mock.patch.object(clusterops.ClusterOps, '_failover_migrate_networks')
    @mock.patch.object(clusterops.ClusterOps, '_nova_failover_server')
    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_failover_same_node(self, mock_get_instance_by_name,
                                mock_nova_failover_server,
                                mock_failover_migrate_networks,
                                mock_wait_pending_instance,
                                recreate_ports_on_failover=False):
        # In some cases, the instances may bounce between hosts. We're testing
        # the case in which the instance is actually returning to the initial
        # host during the time in which we're processing events.
        self.flags(recreate_ports_on_failover=recreate_ports_on_failover,
                   group='hyperv')

        instance = mock_get_instance_by_name.return_value
        old_host = 'old_host'
        new_host = 'new_host'
        instance.host = old_host
        self.clusterops._this_node = old_host
        self._clustutils.get_vm_host.return_value = old_host

        self.clusterops._failover_migrate(mock.sentinel.instance_name,
                                          new_host)

        get_inst_nw_info = self.clusterops._network_api.get_instance_nw_info
        get_inst_nw_info.assert_called_once_with(self.clusterops._context,
                                                 instance)
        mock_nova_failover_server.assert_called_once_with(instance, old_host)
        if recreate_ports_on_failover:
            self.clusterops._vmops.unplug_vifs.assert_called_once_with(
                instance, get_inst_nw_info.return_value)
        else:
            self.clusterops._vmops.unplug_vifs.assert_not_called()
        self.clusterops._vmops.plug_vifs.assert_called_once_with(
            instance, get_inst_nw_info.return_value)
        self._placement.move_compute_node_allocations.assert_not_called()
        mock_failover_migrate_networks.assert_not_called()
        c_handler = self.clusterops._serial_console_ops.start_console_handler
        c_handler.assert_called_once_with(mock.sentinel.instance_name)

    @mock.patch('time.sleep')
    def test_wait_for_pending_instance(self, mock_sleep):
        self._clustutils.get_cluster_group_state_info.side_effect = [
            dict(state=os_win_const.CLUSTER_GROUP_PENDING),
            dict(state=os_win_const.CLUSTER_GROUP_ONLINE)]

        self.clusterops._wait_for_pending_instance(mock.sentinel.instance_name)

        self._clustutils.get_cluster_group_state_info.assert_has_calls(
            [mock.call(mock.sentinel.instance_name)] * 2)
        mock_sleep.assert_called_once_with(2)

    def test_failover_migrate_networks(self):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        fake_source = mock.MagicMock()
        fake_migration = {'source_compute': fake_source,
                          'dest_compute': self.clusterops._this_node}

        self.clusterops._failover_migrate_networks(mock_instance,
                                                    fake_source)

        mock_network_api = self.clusterops._network_api
        calls = [mock.call(self.clusterops._context, mock_instance,
                           self.clusterops._this_node),
                 mock.call(self.clusterops._context, mock_instance,
                           self.clusterops._this_node),
                 mock.call(self.clusterops._context, mock_instance,
                           self.clusterops._this_node),
                 mock.call(self.clusterops._context, mock_instance,
                           fake_source, teardown=True)]
        mock_network_api.setup_networks_on_host.assert_has_calls(calls)
        mock_network_api.migrate_instance_start.assert_called_once_with(
            self.clusterops._context, mock_instance, fake_migration)
        mock_network_api.migrate_instance_finish.assert_called_once_with(
            self.clusterops._context, mock_instance, fake_migration,
            provider_mappings=None)

    @mock.patch.object(objects.Instance, 'get_by_uuid')
    def test_get_instance_by_name(self, mock_get_by_uuid):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        mock_get_by_uuid.return_value = mock_instance
        self.clusterops._instance_map[mock_instance.name] = mock_instance.uuid

        ret = self.clusterops._get_instance_by_name(mock_instance.name)
        self.assertEqual(ret, mock_instance)

    @mock.patch.object(objects.Instance, 'get_by_uuid')
    def test_get_instance_by_name_not_in_cache(self, mock_get_by_uuid):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops._vmutils.get_instance_uuid.return_value = (
            mock_instance.uuid)
        mock_get_by_uuid.return_value = mock_instance

        ret = self.clusterops._get_instance_by_name(mock_instance.name)
        self.assertEqual(ret, mock_instance)
        self.assertEqual(mock_instance.uuid,
                         self.clusterops._instance_map[mock_instance.name])

    @mock.patch.object(objects.Instance, 'get_by_uuid')
    def test_get_instance_by_name_not_update_map(self, mock_get_by_uuid):
        mock_instance = fake_instance.fake_instance_obj(self.context)
        self.clusterops._vmutils.get_instance_uuid.side_effect = (
            os_win_exc.HyperVVMNotFoundException(vm_name=mock_instance.name))
        self.clusterops._update_instance_map = mock.MagicMock()
        self.clusterops._instance_map = mock.MagicMock()
        self.clusterops._instance_map.get.side_effect = [None,
                                                          mock_instance.uuid]
        mock_get_by_uuid.return_value = mock_instance

        ret = self.clusterops._get_instance_by_name(mock_instance.name)
        self.assertEqual(ret, mock_instance)
        self.clusterops._update_instance_map.assert_called_with()

    @mock.patch.object(clusterops.ClusterOps, '_get_nova_instances')
    def test_update_instance_map(self, mock_get_instances):
        mock_instance = mock.MagicMock(uuid=mock.sentinel.uuid)
        mock_instance.configure_mock(name=mock.sentinel.name)
        mock_get_instances.return_value = [mock_instance]

        self.clusterops._update_instance_map()

        self.assertEqual(mock.sentinel.uuid,
                         self.clusterops._instance_map[mock.sentinel.name])

    @ddt.data({'instance_uuids': None},
              {'instance_uuids': []},
              {'instance_uuids': mock.sentinel.uuid})
    @ddt.unpack
    @mock.patch.object(clusterops.objects.InstanceList, 'get_by_filters')
    def test_get_nova_instances(self, mock_get_by_filters, instance_uuids):
        instances = self.clusterops._get_nova_instances(
            instance_uuids=instance_uuids)

        self.assertEqual(mock_get_by_filters.return_value, instances)
        expected_attrs = ['id', 'uuid', 'name']
        expected_filters = {'deleted': False}
        if instance_uuids is not None:
            expected_filters['uuid'] = instance_uuids
        mock_get_by_filters.assert_called_once_with(
            self.clusterops._context, expected_filters,
            expected_attrs=expected_attrs)

    @mock.patch.object(clusterops.block_device, 'DriverVolumeBlockDevice')
    @mock.patch.object(clusterops.objects.BlockDeviceMappingList,
                       'get_by_instance_uuid')
    def test_get_instance_block_device_mappings(self, mock_get_by_uuid,
                                                mock_DriverVBD):
        mock_get_by_uuid.return_value = [mock.sentinel.bdm]
        mock_instance = mock.MagicMock()

        bdms = self.clusterops._get_instance_block_device_mappings(
            mock_instance)

        self.assertEqual([mock_DriverVBD.return_value], bdms)
        mock_get_by_uuid.assert_called_once_with(self.clusterops._context,
                                                 mock_instance.uuid)
        mock_DriverVBD.assert_called_once_with(mock.sentinel.bdm)

    def test_nova_failover_server(self):
        mock_instance = mock.MagicMock(vm_state=vm_states.ERROR,
                                       power_state=power_state.NOSTATE)

        self.clusterops._nova_failover_server(mock_instance,
                                               mock.sentinel.host)

        self.assertEqual(vm_states.ACTIVE, mock_instance.vm_state)
        self.assertEqual(power_state.RUNNING, mock_instance.power_state)
        self.assertEqual(mock.sentinel.host, mock_instance.host)
        self.assertEqual(mock.sentinel.host, mock_instance.node)
        mock_instance.save.assert_called_once_with(expected_task_state=[None])

    @mock.patch.object(clusterops.ClusterOps, '_get_instance_by_name')
    def test_instance_state_change_callback(self, mock_get_instance_by_name):
        event = mock.Mock(transition=virtevent.EVENT_LIFECYCLE_STARTED)
        mock_instance = mock_get_instance_by_name.return_value

        self.clusterops.instance_state_change_callback(event)

        mock_get_instance_by_name.assert_called_once_with(event.name)
        self._network_api.get_instance_nw_info.assert_called_once_with(
            self.context, mock_instance)
        self.clusterops._vmops.plug_vifs.assert_called_once_with(
            mock_instance,
            self._network_api.get_instance_nw_info.return_value)



#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid

from nova import objects
from nova.objects import fields
import six


def fake_db_instance(**updates):
    flavorinfo = None
    db_instance = {
        'id': 1,
        'deleted': False,
        'uuid': str(uuid.uuid4()),
        'user_id': 'fake-user',
        'project_id': 'fake-project',
        'host': 'fake-host',
        'created_at': datetime.datetime(1955, 11, 5),
        'pci_devices': [],
        'security_groups': [],
        'metadata': {},
        'system_metadata': {},
        'root_gb': 0,
        'ephemeral_gb': 0,
        'extra': {'pci_requests': None,
                  'flavor': flavorinfo,
                  'numa_topology': None,
                  'vcpu_model': None,
                  'trusted_certs': None,
                 },
        'tags': [],
        'services': []
        }

    for name, field in six.iteritems(objects.Instance.fields):
        if name in db_instance:
            continue
        if field.nullable:
            db_instance[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_instance[name] = field.default
        elif name in ['flavor', 'ec2_ids', 'keypairs']:
            pass
        else:
            raise Exception('fake_db_instance needs help with %s' % name)

    if updates:
        db_instance.update(updates)
    return db_instance


def fake_instance_obj(context='fake-context', **updates):
    expected_attrs = updates.pop('expected_attrs', None)
    flavor = objects.Flavor(id=1, name='flavor1',
                            memory_mb=256, vcpus=1,
                            root_gb=1, ephemeral_gb=1,
                            flavorid='1',
                            swap=0, rxtx_factor=1.0,
                            vcpu_weight=1,
                            disabled=False,
                            is_public=True,
                            extra_specs={},
                            projects=[])
    flavor.obj_reset_changes()
    inst = objects.Instance._from_db_object(context,
               objects.Instance(), fake_db_instance(**updates),
               expected_attrs=expected_attrs)
    inst.flavor = flavor
    inst.old_flavor = flavor.obj_clone()
    inst.obj_reset_changes()
    return inst



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Management class for cluster live migration VM operations."""

from nova.compute import vm_states
from nova import exception
from os_win import constants as os_win_const
from os_win import utilsfactory
from oslo_log import log as logging

from compute_hyperv.i18n import _
import compute_hyperv.nova.conf
from compute_hyperv.nova import livemigrationops

CONF = compute_hyperv.nova.conf.CONF
LOG = logging.getLogger(__name__)


class ClusterLiveMigrationOps(livemigrationops.LiveMigrationOps):
    def __init__(self):
        super(ClusterLiveMigrationOps, self).__init__()
        self._clustutils = utilsfactory.get_clusterutils()

    def is_instance_clustered(self, instance_name):
        return self._clustutils.vm_exists(instance_name)

    def live_migration(self, context, instance_ref, dest, post_method,
                       recover_method, block_migration=False,
                       migrate_data=None):
        LOG.debug("live_migration called.", instance=instance_ref)
        instance_name = instance_ref.name
        clustered = self.is_instance_clustered(instance_name)
        node_names = [node.upper() for node in
                      self._clustutils.get_cluster_node_names()]

        if dest.upper() not in node_names or not clustered:
            # destination is not in same cluster or instance not clustered.
            # do a normal live migration.
            if clustered:
                # remove VM from cluster before proceding to a normal live
                # migration.
                self._clustutils.delete(instance_name)
            super(ClusterLiveMigrationOps, self).live_migration(
                context, instance_ref, dest, post_method, recover_method,
                block_migration, migrate_data)
            return
        elif self._clustutils.get_vm_host(
                instance_name).upper() == dest.upper():
            # VM is already migrated. Do nothing.
            # this can happen when the VM has been failovered.
            return

        # destination is in the same cluster.
        # perform a clustered live migration.
        try:
            self._clustutils.live_migrate_vm(
                instance_name,
                dest,
                CONF.hyperv.instance_live_migration_timeout)
        except Exception:
            LOG.exception("Live migration failed. Attempting rollback.",
                          instance=instance_ref)
            # The recover method will update the migration state.
            # We won't error out if we manage to recover the instance,
            # which would otherwise end up in error state.
            self._check_failed_instance_migration(
                instance_ref,
                expected_state=os_win_const.CLUSTER_GROUP_ONLINE)

            recover_method(context, instance_ref, dest, migrate_data)
            return

        LOG.debug("Calling live migration post_method for instance.",
                  instance=instance_ref)
        post_method(context, instance_ref, dest,
                    block_migration, migrate_data)

    def _check_failed_instance_migration(self, instance, expected_state):
        # After a failed migration, we expect the instance to be on the
        # source node, having its initial state and not have any queued
        # migrations. Otherwise, we treat it as a critical error and set
        # it to 'error' state to avoid inconsistencies.
        state_info = self._clustutils.get_cluster_group_state_info(
            instance.name)
        node_name = self._clustutils.get_node_name()

        if (state_info['owner_node'].lower() != node_name.lower() or
                state_info['state'] != expected_state or
                state_info['migration_queued']):
            instance.vm_state = vm_states.ERROR
            instance.save()

            raise exception.InstanceInvalidState(
                _("Instance %(instance_name)s reached an inconsistent state "
                  "after a failed migration attempt. Setting the instance to "
                  "'error' state. Instance state info: %(state_info)s.") %
                dict(instance_name=instance.name,
                     state_info=state_info))

    def pre_live_migration(self, context, instance, block_device_info,
                           network_info):
        if self.is_instance_clustered(instance.name):
            self._volumeops.connect_volumes(block_device_info)
        else:
            super(ClusterLiveMigrationOps, self).pre_live_migration(
                context, instance, block_device_info, network_info)

    def post_live_migration(self, context, instance, block_device_info,
                            migrate_data):
        if not self.is_instance_clustered(instance.name):
            super(ClusterLiveMigrationOps, self).post_live_migration(
                context, instance, block_device_info, migrate_data)



# Copyright 2016 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for the Hyper-V Cluster Driver."""

from unittest import mock

from nova import safe_utils
from nova.virt import driver as nova_base_driver

from compute_hyperv.nova.cluster import driver
from compute_hyperv.nova import driver as base_driver
from compute_hyperv.tests.unit import test_base


class HyperVClusterTestCase(test_base.HyperVBaseTestCase):

    _autospec_classes = [
        driver.clusterops.ClusterOps,
        base_driver.eventhandler.InstanceEventHandler,
        base_driver.hostops.api.API,
        driver.livemigrationops.ClusterLiveMigrationOps,
    ]

    @mock.patch.object(base_driver.HyperVDriver,
                       '_check_minimum_windows_version')
    def setUp(self, mock_check_minimum_windows_version):
        super(HyperVClusterTestCase, self).setUp()

        self.context = 'context'
        self.driver = driver.HyperVClusterDriver(mock.sentinel.virtapi)

    def test_public_api_signatures(self):
        driver_methods = dict(driver.HyperVClusterDriver.__dict__,
                              **base_driver.HyperVDriver.__dict__)

        for attr in driver_methods:
            class_member = getattr(driver.HyperVClusterDriver, attr)
            if callable(class_member):
                mocked_method = mock.patch.object(
                    driver.HyperVClusterDriver, attr,
                    safe_utils.get_wrapped_function(class_member))
                mocked_method.start()
                self.addCleanup(mocked_method.stop)

        self.assertPublicAPISignatures(nova_base_driver.ComputeDriver,
                                       driver.HyperVClusterDriver)

    def test_set_event_handler_callbacks(self):
        self.driver._set_event_handler_callbacks()

        self.driver._event_handler.add_callback.assert_has_calls(
            [mock.call(self.driver.emit_event),
             mock.call(self.driver._vmops.instance_state_change_callback),
             mock.call(self.driver._clops.instance_state_change_callback)])

    @mock.patch.object(base_driver.HyperVDriver, 'spawn')
    def test_spawn(self, mock_superclass_spawn):
        self.driver.spawn(self.context, mock.sentinel.fake_instance,
                          mock.sentinel.image_meta,
                          mock.sentinel.injected_files,
                          mock.sentinel.admin_pass,
                          mock.sentinel.allocations,
                          mock.sentinel.network_info,
                          mock.sentinel.block_dev_info,
                          mock.sentinel.power_on,
                          mock.sentinel.accel_info)

        mock_superclass_spawn.assert_called_once_with(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.image_meta, mock.sentinel.injected_files,
            mock.sentinel.admin_pass, mock.sentinel.allocations,
            mock.sentinel.network_info, mock.sentinel.block_dev_info,
            mock.sentinel.power_on)
        self.driver._clops.add_to_cluster.assert_called_once_with(
            mock.sentinel.fake_instance)

    @mock.patch.object(base_driver.HyperVDriver, 'destroy')
    def test_destroy(self, mock_superclass_destroy):
        self.driver.destroy(self.context, mock.sentinel.fake_instance,
                            mock.sentinel.network_info,
                            mock.sentinel.block_dev_info,
                            mock.sentinel.destroy_disks)

        mock_superclass_destroy.assert_called_once_with(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.network_info, mock.sentinel.block_dev_info,
            mock.sentinel.destroy_disks)

    @mock.patch.object(base_driver.HyperVDriver, 'migrate_disk_and_power_off')
    def test_migrate_disk_and_power_off(self, mock_superclass_migrate):
        disk_info = self.driver.migrate_disk_and_power_off(
            self.context,
            mock.sentinel.fake_instance,
            mock.sentinel.destination,
            mock.sentinel.flavor,
            mock.sentinel.network_info,
            mock.sentinel.block_dev_info,
            mock.sentinel.timeout,
            mock.sentinel.retry_interval)

        self.assertEqual(mock_superclass_migrate.return_value, disk_info)
        self.driver._clops.remove_from_cluster.assert_called_once_with(
            mock.sentinel.fake_instance)
        mock_superclass_migrate.assert_called_once_with(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.destination, mock.sentinel.flavor,
            mock.sentinel.network_info, mock.sentinel.block_dev_info,
            mock.sentinel.timeout, mock.sentinel.retry_interval)

    @mock.patch.object(base_driver.HyperVDriver, 'finish_migration')
    def test_finish_migration(self, mock_superclass_finish_migration):
        self.driver.finish_migration(self.context,
                                     mock.sentinel.migration,
                                     mock.sentinel.fake_instance,
                                     mock.sentinel.disk_info,
                                     mock.sentinel.network_info,
                                     mock.sentinel.image_meta,
                                     mock.sentinel.resize_instance,
                                     mock.sentinel.allocations,
                                     mock.sentinel.block_dev_info,
                                     mock.sentinel.power_on)
        mock_superclass_finish_migration.assert_called_once_with(
            self.context, mock.sentinel.migration, mock.sentinel.fake_instance,
            mock.sentinel.disk_info, mock.sentinel.network_info,
            mock.sentinel.image_meta, mock.sentinel.resize_instance,
            mock.sentinel.allocations, mock.sentinel.block_dev_info,
            mock.sentinel.power_on)
        self.driver._clops.add_to_cluster.assert_called_once_with(
            mock.sentinel.fake_instance)

    @mock.patch.object(base_driver.HyperVDriver, 'finish_revert_migration')
    def test_finish_revert_migration(self, mock_superclass_finish_rev_migr):
        self.driver.finish_revert_migration(self.context,
                                            mock.sentinel.fake_instance,
                                            mock.sentinel.network_info,
                                            mock.sentinel.migration,
                                            mock.sentinel.block_dev_info,
                                            mock.sentinel.power_on)
        mock_superclass_finish_rev_migr.assert_called_once_with(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.network_info, mock.sentinel.migration,
            mock.sentinel.block_dev_info, mock.sentinel.power_on)
        self.driver._clops.add_to_cluster.assert_called_once_with(
            mock.sentinel.fake_instance)

    @mock.patch.object(driver.HyperVClusterDriver, 'unplug_vifs')
    def test_rollback_live_migration_at_destination_clustered(
            self, mock_unplug_vifs):
        mock_is_clustered = self.driver._livemigrationops.is_instance_clustered
        mock_instance = mock.Mock()
        self.driver.rollback_live_migration_at_destination(
            self.context, mock_instance, mock.sentinel.network_info,
            mock.sentinel.block_dev_info, mock.sentinel.destroy_disks,
            mock.sentinel.migrate_data)

        mock_is_clustered.assert_called_once_with(mock_instance.name)
        mock_unplug_vifs.assert_called_once_with(
            mock_instance, mock.sentinel.network_info)

    @mock.patch.object(base_driver.HyperVDriver,
                       'rollback_live_migration_at_destination')
    def test_rollback_live_migration_at_destination(self,
                                                    mock_superclass_rollback):
        mock_is_clustered = self.driver._livemigrationops.is_instance_clustered
        mock_is_clustered.return_value = False
        mock_instance = mock.Mock()
        self.driver.rollback_live_migration_at_destination(
            self.context, mock_instance, mock.sentinel.network_info,
            mock.sentinel.block_dev_info, mock.sentinel.destroy_disks,
            mock.sentinel.migrate_data)

        mock_is_clustered.assert_called_once_with(mock_instance.name)
        mock_superclass_rollback.assert_called_once_with(
            self.context, mock_instance, mock.sentinel.network_info,
            mock.sentinel.block_dev_info, mock.sentinel.destroy_disks,
            mock.sentinel.migrate_data)

    @mock.patch.object(base_driver.HyperVDriver,
                       'post_live_migration_at_destination')
    def test_post_live_migration_at_destination(self, mock_superclass_post):
        self.driver.post_live_migration_at_destination(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.network_info, mock.sentinel.block_migration,
            mock.sentinel.block_dev_info)

        self.driver._clops.post_migration.assert_called_once_with(
            mock.sentinel.fake_instance)
        mock_superclass_post.assert_called_once_with(
            self.context, mock.sentinel.fake_instance,
            mock.sentinel.network_info, mock.sentinel.block_migration,
            mock.sentinel.block_dev_info)



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

import ddt
from nova import exception
from nova.network import model
from os_win import constants as os_win_const

import compute_hyperv.nova.conf
from compute_hyperv.nova import vif
from compute_hyperv.tests.unit import test_base


CONF = compute_hyperv.nova.conf.CONF


class HyperVNovaNetworkVIFPluginTestCase(test_base.HyperVBaseTestCase):
    def setUp(self):
        super(HyperVNovaNetworkVIFPluginTestCase, self).setUp()
        self.vif_driver = vif.HyperVNovaNetworkVIFPlugin()

    def test_plug(self):
        self.flags(vswitch_name='fake_vswitch_name', group='hyperv')
        fake_vif = {'id': mock.sentinel.fake_id}

        self.vif_driver.plug(mock.sentinel.instance, fake_vif)
        netutils = self.vif_driver._netutils
        netutils.connect_vnic_to_vswitch.assert_called_once_with(
            'fake_vswitch_name', mock.sentinel.fake_id)


@ddt.ddt
class HyperVVIFDriverTestCase(test_base.HyperVBaseTestCase):
    def setUp(self):
        super(HyperVVIFDriverTestCase, self).setUp()
        self.vif_driver = vif.HyperVVIFDriver()
        self.vif_driver._vif_plugin = mock.MagicMock()

        self._netutils = self.vif_driver._netutils
        self._vmutils = self.vif_driver._vmutils
        self._metricsutils = self.vif_driver._metricsutils

    def test_plug(self):
        vif = {'type': model.VIF_TYPE_HYPERV}
        self.vif_driver.plug(mock.sentinel.instance, vif)

        self.vif_driver._vif_plugin.plug.assert_called_once_with(
            mock.sentinel.instance, vif)

    @mock.patch.object(vif, 'os_vif')
    @mock.patch.object(vif.HyperVVIFDriver, 'enable_metrics')
    @mock.patch.object(vif.os_vif_util, 'nova_to_osvif_instance')
    @mock.patch.object(vif.os_vif_util, 'nova_to_osvif_vif')
    def test_plug_ovs(self, mock_nova_to_osvif_vif,
                      mock_nova_to_osvif_instance,
                      mock_enable_metrics, mock_os_vif):
        vif = {'type': model.VIF_TYPE_OVS}
        self.assertRaises(exception.VirtualInterfacePlugException,
                          self.vif_driver.plug,
                          mock.sentinel.instance, vif)

    @ddt.data(True, False)
    def test_enable_metrics(self, vm_running):
        state = (os_win_const.HYPERV_VM_STATE_ENABLED if vm_running
                 else os_win_const.HYPERV_VM_STATE_DISABLED)
        self._vmutils.get_vm_state.return_value = state

        enable_metrics = self._metricsutils.enable_port_metrics_collection

        self.vif_driver.enable_metrics(mock.sentinel.instance_name,
                                       mock.sentinel.vif_id)

        self._vmutils.get_vm_state.assert_called_once_with(
            mock.sentinel.instance_name)
        if vm_running:
            enable_metrics.assert_called_once_with(mock.sentinel.vif_id)
        else:
            enable_metrics.assert_not_called()

    def test_plug_type_unknown(self):
        vif = {'type': mock.sentinel.vif_type}
        self.assertRaises(exception.VirtualInterfacePlugException,
                          self.vif_driver.plug,
                          mock.sentinel.instance, vif)

    def test_unplug(self):
        vif = {'type': model.VIF_TYPE_HYPERV}
        self.vif_driver.unplug(mock.sentinel.instance, vif)

        self.vif_driver._vif_plugin.unplug.assert_called_once_with(
            mock.sentinel.instance, vif)

    @mock.patch.object(vif, 'os_vif')
    @mock.patch.object(vif.os_vif_util, 'nova_to_osvif_instance')
    @mock.patch.object(vif.os_vif_util, 'nova_to_osvif_vif')
    def test_unplug_ovs(self, mock_nova_to_osvif_vif,
                        mock_nova_to_osvif_instance, mock_os_vif):
        vif = {'type': model.VIF_TYPE_OVS}
        self.vif_driver.unplug(mock.sentinel.instance, vif)

        mock_nova_to_osvif_vif.assert_called_once_with(vif)
        mock_nova_to_osvif_instance.assert_called_once_with(
            mock.sentinel.instance)
        mock_os_vif.unplug.assert_called_once_with(
            mock_nova_to_osvif_vif.return_value,
            mock_nova_to_osvif_instance.return_value)

    def test_unplug_type_unknown(self):
        vif = {'type': mock.sentinel.vif_type}
        self.assertRaises(exception.VirtualInterfaceUnplugException,
                          self.vif_driver.unplug,
                          mock.sentinel.instance, vif)



# Copyright (c) 2016 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""A Hyper-V Cluster Nova Compute driver."""

from compute_hyperv.nova.cluster import clusterops
from compute_hyperv.nova.cluster import livemigrationops
from compute_hyperv.nova.cluster import volumeops
from compute_hyperv.nova import driver


class HyperVClusterDriver(driver.HyperVDriver):
    use_coordination = True

    def __init__(self, virtapi):
        super(HyperVClusterDriver, self).__init__(virtapi)

        self._clops = clusterops.ClusterOps()
        self._livemigrationops = livemigrationops.ClusterLiveMigrationOps()
        self._volumeops = volumeops.ClusterVolumeOps()

        self._clops.start_failover_listener_daemon()
        self._clops.reclaim_failovered_instances()

    def _set_event_handler_callbacks(self):
        super(HyperVClusterDriver, self)._set_event_handler_callbacks()

        self._event_handler.add_callback(
            self._clops.instance_state_change_callback)

    def spawn(self, context, instance, image_meta, injected_files,
              admin_password, allocations, network_info=None,
              block_device_info=None, power_on=True, accel_info=None):
        super(HyperVClusterDriver, self).spawn(
            context, instance, image_meta, injected_files, admin_password,
            allocations, network_info, block_device_info, power_on)
        self._clops.add_to_cluster(instance)

    def destroy(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, destroy_secrets=True):
        self._clops.remove_from_cluster(instance)
        super(HyperVClusterDriver, self).destroy(
            context, instance, network_info, block_device_info,
            destroy_disks)

    def migrate_disk_and_power_off(self, context, instance, dest,
                                   flavor, network_info,
                                   block_device_info=None,
                                   timeout=0, retry_interval=0):
        self._clops.remove_from_cluster(instance)
        return super(HyperVClusterDriver, self).migrate_disk_and_power_off(
            context, instance, dest, flavor, network_info,
            block_device_info, timeout, retry_interval)

    def finish_migration(self, context, migration, instance, disk_info,
                         network_info, image_meta, resize_instance,
                         allocations, block_device_info=None, power_on=True):
        super(HyperVClusterDriver, self).finish_migration(
            context, migration, instance, disk_info, network_info,
            image_meta, resize_instance, allocations,
            block_device_info, power_on)
        self._clops.add_to_cluster(instance)

    def finish_revert_migration(self, context, instance, network_info,
                                migration, block_device_info=None,
                                power_on=True):
        super(HyperVClusterDriver, self).finish_revert_migration(
            context, instance, network_info, migration,
            block_device_info, power_on)
        self._clops.add_to_cluster(instance)

    def rollback_live_migration_at_destination(self, context, instance,
                                               network_info,
                                               block_device_info,
                                               destroy_disks=True,
                                               migrate_data=None):
        if self._livemigrationops.is_instance_clustered(instance.name):
            self.unplug_vifs(instance, network_info)
        else:
            super(HyperVClusterDriver,
                  self).rollback_live_migration_at_destination(
                    context, instance, network_info, block_device_info,
                    destroy_disks, migrate_data)

    def post_live_migration_at_destination(self, context, instance,
                                           network_info,
                                           block_migration=False,
                                           block_device_info=None):
        self._clops.post_migration(instance)
        super(HyperVClusterDriver, self).post_live_migration_at_destination(
            context, instance, network_info,
            block_migration, block_device_info)



# Copyright 2012 Pedro Navarro Perez
# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Management class for Storage-related functions (attach, detach, etc).
"""
import inspect
import os
import time

from nova.compute import task_states
from nova import exception
from nova import objects
from nova import utils
from nova.virt import block_device as driver_block_device
from nova.virt import driver
from nova.volume import cinder
from os_brick.initiator import connector
from os_win import constants as os_win_const
from os_win import utilsfactory
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import strutils

from compute_hyperv.i18n import _
import compute_hyperv.nova.conf
from compute_hyperv.nova import constants
from compute_hyperv.nova import pathutils

LOG = logging.getLogger(__name__)

CONF = compute_hyperv.nova.conf.CONF


def volume_snapshot_lock(f):
    """Synchronizes volume snapshot related operations.

    The locks will be applied on a per-instance basis. The decorated method
    must accept an instance object.
    """
    def inner(*args, **kwargs):
        all_args = inspect.getcallargs(f, *args, **kwargs)
        instance = all_args['instance']

        lock_name = "volume-snapshot-%s" % instance.name

        @utils.synchronized(lock_name)
        def synchronized():
            return f(*args, **kwargs)

        return synchronized()
    return inner


class VolumeOps(object):
    """Management class for Volume-related tasks
    """

    def __init__(self):
        self._volume_api = cinder.API()
        self._vmops_prop = None
        self._block_dev_man_prop = None

        self._vmutils = utilsfactory.get_vmutils()
        self._default_root_device = 'vda'

        self._load_volume_drivers()

    def _load_volume_drivers(self):
        self.volume_drivers = {
            constants.STORAGE_PROTOCOL_SMBFS: SMBFSVolumeDriver(),
            constants.STORAGE_PROTOCOL_ISCSI: ISCSIVolumeDriver(),
            constants.STORAGE_PROTOCOL_FC: FCVolumeDriver(),
            constants.STORAGE_PROTOCOL_RBD: RBDVolumeDriver()}

    @property
    def _vmops(self):
        # We have to avoid a circular dependency.
        if not self._vmops_prop:
            self._vmops_prop = importutils.import_class(
                'compute_hyperv.nova.vmops.VMOps')()
        return self._vmops_prop

    @property
    def _block_dev_man(self):
        if not self._block_dev_man_prop:
            self._block_dev_man_prop = importutils.import_class(
                'compute_hyperv.nova.block_device_manager.'
                'BlockDeviceInfoManager')()
        return self._block_dev_man_prop

    def _get_volume_driver(self, connection_info):
        driver_type = connection_info.get('driver_volume_type')
        if driver_type not in self.volume_drivers:
            raise exception.VolumeDriverNotFound(driver_type=driver_type)
        return self.volume_drivers[driver_type]

    def validate_host_configuration(self):
        for protocol, volume_driver in self.volume_drivers.items():
            try:
                volume_driver.validate_host_configuration()
            except exception.ValidationError as ex:
                LOG.warning(
                    "Volume driver %(protocol)s reported a validation "
                    "error. Attaching such volumes will probably fail. "
                    "Error message: %(err_msg)s.",
                    dict(protocol=protocol, err_msg=ex.message))

    def attach_volumes(self, context, volumes, instance):
        for vol in volumes:
            self.attach_volume(context, vol['connection_info'], instance)

    def disconnect_volumes(self, block_device_info):
        mapping = driver.block_device_info_get_mapping(block_device_info)
        for vol in mapping:
            self.disconnect_volume(vol['connection_info'])

    def attach_volume(self, context, connection_info, instance,
                      disk_bus=constants.CTRL_TYPE_SCSI,
                      update_device_metadata=False):
        tries_left = CONF.hyperv.volume_attach_retry_count + 1

        while tries_left:
            try:
                self._attach_volume(context,
                                    connection_info,
                                    instance,
                                    disk_bus,
                                    update_device_metadata)
                break
            except Exception as ex:
                tries_left -= 1
                if not tries_left:
                    LOG.exception(
                        "Failed to attach volume %(connection_info)s "
                        "to instance %(instance_name)s. ",
                        {'connection_info': strutils.mask_dict_password(
                             connection_info),
                         'instance_name': instance.name})

                    # We're requesting a detach as the disk may have
                    # been attached to the instance but one of the
                    # post-attach operations failed.
                    self.detach_volume(context,
                                       connection_info,
                                       instance,
                                       update_device_metadata)
                    raise exception.VolumeAttachFailed(
                        volume_id=connection_info['serial'],
                        reason=ex)
                else:
                    LOG.warning(
                        "Failed to attach volume %(connection_info)s "
                        "to instance %(instance_name)s. "
                        "Tries left: %(tries_left)s.",
                        {'connection_info': strutils.mask_dict_password(
                             connection_info),
                         'instance_name': instance.name,
                         'tries_left': tries_left})

                    time.sleep(CONF.hyperv.volume_attach_retry_interval)

    def _attach_volume(self, context, connection_info, instance,
                       disk_bus=constants.CTRL_TYPE_SCSI,
                       update_device_metadata=False):
        LOG.debug(
            "Attaching volume: %(connection_info)s to %(instance_name)s",
            {'connection_info': strutils.mask_dict_password(connection_info),
             'instance_name': instance.name})
        volume_driver = self._get_volume_driver(connection_info)
        volume_driver.attach_volume(connection_info,
                                    instance.name,
                                    disk_bus)

        if update_device_metadata:
            # When attaching volumes to already existing instances,
            # the connection info passed to the driver is not saved
            # yet within the BDM table.
            self._block_dev_man.set_volume_bdm_connection_info(
                context, instance, connection_info)
            self._vmops.update_device_metadata(
                context, instance)

        qos_specs = connection_info['data'].get('qos_specs') or {}
        if qos_specs:
            volume_driver.set_disk_qos_specs(connection_info,
                                             qos_specs)

    def disconnect_volume(self, connection_info):
        volume_driver = self._get_volume_driver(connection_info)
        volume_driver.disconnect_volume(connection_info)

    def detach_volume(self, context, connection_info, instance,
                      update_device_metadata=False):
        LOG.debug("Detaching volume: %(connection_info)s "
                  "from %(instance_name)s",
                  {'connection_info': strutils.mask_dict_password(
                      connection_info),
                   'instance_name': instance.name})
        volume_driver = self._get_volume_driver(connection_info)
        volume_driver.detach_volume(connection_info, instance.name)
        volume_driver.disconnect_volume(connection_info)

        if update_device_metadata:
            self._vmops.update_device_metadata(context, instance)

    def fix_instance_volume_disk_paths(self, instance_name, block_device_info):
        # Mapping containing the current disk paths for each volume.
        actual_disk_mapping = self.get_disk_path_mapping(block_device_info)
        if not actual_disk_mapping:
            return

        # Mapping containing virtual disk resource path and the physical
        # disk path for each volume serial number. The physical path
        # associated with this resource may not be the right one,
        # as physical disk paths can get swapped after host reboots.
        vm_disk_mapping = self._vmutils.get_vm_physical_disk_mapping(
            instance_name)

        for serial, vm_disk in vm_disk_mapping.items():
            actual_disk_path = actual_disk_mapping[serial]
            if vm_disk['mounted_disk_path'] != actual_disk_path:
                self._vmutils.set_disk_host_res(vm_disk['resource_path'],
                                                actual_disk_path)

    def get_volume_connector(self):
        # NOTE(lpetrut): the Windows os-brick connectors
        # do not use a root helper.
        conn = connector.get_connector_properties(
            root_helper=None,
            my_ip=CONF.my_block_storage_ip,
            multipath=CONF.hyperv.use_multipath_io,
            enforce_multipath=True,
            host=CONF.host)
        return conn

    def connect_volumes(self, block_device_info):
        mapping = driver.block_device_info_get_mapping(block_device_info)
        for vol in mapping:
            connection_info = vol['connection_info']
            volume_driver = self._get_volume_driver(connection_info)
            volume_driver.connect_volume(connection_info)

    def get_disk_path_mapping(self, block_device_info, block_dev_only=False):
        block_mapping = driver.block_device_info_get_mapping(block_device_info)
        disk_path_mapping = {}
        for vol in block_mapping:
            connection_info = vol['connection_info']
            disk_serial = connection_info['serial']

            volume_driver = self._get_volume_driver(connection_info)
            if block_dev_only and not volume_driver._is_block_dev:
                continue

            disk_path = volume_driver.get_disk_resource_path(connection_info)
            disk_path_mapping[disk_serial] = disk_path
        return disk_path_mapping

    def get_disk_resource_path(self, connection_info):
        volume_driver = self._get_volume_driver(connection_info)
        return volume_driver.get_disk_resource_path(connection_info)

    @staticmethod
    def bytes_per_sec_to_iops(no_bytes):
        # Hyper-v uses normalized IOPS (8 KB increments)
        # as IOPS allocation units.
        return (
            (no_bytes + constants.IOPS_BASE_SIZE - 1) //
            constants.IOPS_BASE_SIZE)

    @staticmethod
    def validate_qos_specs(qos_specs, supported_qos_specs):
        unsupported_specs = set(qos_specs.keys()).difference(
            supported_qos_specs)
        if unsupported_specs:
            LOG.warning('Got unsupported QoS specs: '
                        '%(unsupported_specs)s. '
                        'Supported qos specs: %(supported_qos_specs)s',
                        {'unsupported_specs': unsupported_specs,
                         'supported_qos_specs': supported_qos_specs})

    @volume_snapshot_lock
    def volume_snapshot_create(self, context, instance, volume_id,
                               create_info):
        LOG.debug("Creating snapshot for volume %(volume_id)s on instance "
                  "%(instance_name)s with create info %(create_info)s",
                  {"volume_id": volume_id,
                   "instance_name": instance.name,
                   "create_info": create_info})
        snapshot_id = create_info['snapshot_id']

        snapshot_failed = False
        try:
            instance.task_state = task_states.IMAGE_SNAPSHOT_PENDING
            instance.save(expected_task_state=[None])

            bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
                context, volume_id, instance.uuid)
            driver_bdm = driver_block_device.convert_volume(bdm)
            connection_info = driver_bdm['connection_info']

            volume_driver = self._get_volume_driver(connection_info)
            volume_driver.create_snapshot(connection_info, instance,
                                          create_info)

            # The volume driver is expected to
            # update the connection info.
            driver_bdm.save()
        except Exception:
            with excutils.save_and_reraise_exception():
                snapshot_failed = True

                err_msg = ('Error occurred while snapshotting volume. '
                           'sending error status to Cinder.')
                LOG.exception(err_msg,
                              instance=instance)
        finally:
            instance.task_state = None
            instance.save(
                expected_task_state=[task_states.IMAGE_SNAPSHOT_PENDING])

            snapshot_status = 'error' if snapshot_failed else 'creating'
            self._volume_api.update_snapshot_status(
                context, snapshot_id, snapshot_status)

    @volume_snapshot_lock
    def volume_snapshot_delete(self, context, instance, volume_id,
                               snapshot_id, delete_info):
        LOG.debug("Deleting snapshot for volume %(volume_id)s on instance "
                  "%(instance_name)s with delete info %(delete_info)s",
                  {"volume_id": volume_id,
                   "instance_name": instance.name,
                   "delete_info": delete_info})

        snapshot_delete_failed = False
        try:
            instance.task_state = task_states.IMAGE_SNAPSHOT_PENDING
            instance.save(expected_task_state=[None])

            bdm = objects.BlockDeviceMapping.get_by_volume_and_instance(
                context, volume_id, instance.uuid)
            driver_bdm = driver_block_device.convert_volume(bdm)
            connection_info = driver_bdm['connection_info']

            volume_driver = self._get_volume_driver(connection_info)
            volume_driver.delete_snapshot(connection_info, instance,
                                          delete_info)

            # The volume driver is expected to
            # update the connection info.
            driver_bdm.save()
        except Exception:
            with excutils.save_and_reraise_exception():
                snapshot_delete_failed = True

                err_msg = ('Error occurred while deleting volume '
                           'snapshot. Sending error status to Cinder.')
                LOG.exception(err_msg,
                              instance=instance)
        finally:
            instance.task_state = None
            instance.save(
                expected_task_state=[task_states.IMAGE_SNAPSHOT_PENDING])

            snapshot_status = ('error_deleting'
                               if snapshot_delete_failed else 'deleting')
            self._volume_api.update_snapshot_status(
                context, snapshot_id, snapshot_status)

    def get_disk_attachment_info(self, connection_info):
        volume_driver = self._get_volume_driver(connection_info)
        return volume_driver.get_disk_attachment_info(connection_info)

    def extend_volume(self, connection_info):
        volume_driver = self._get_volume_driver(connection_info)
        return volume_driver.extend_volume(connection_info)


class BaseVolumeDriver(object):
    _is_block_dev = True
    _protocol = None
    _extra_connector_args = {}

    def __init__(self):
        self._conn = None
        self._diskutils = utilsfactory.get_diskutils()
        self._vmutils = utilsfactory.get_vmutils()
        self._migrutils = utilsfactory.get_migrationutils()
        self._metricsutils = utilsfactory.get_metricsutils()

    @property
    def _connector(self):
        if not self._conn:
            scan_attempts = CONF.hyperv.mounted_disk_query_retry_count
            scan_interval = CONF.hyperv.mounted_disk_query_retry_interval

            self._conn = connector.InitiatorConnector.factory(
                protocol=self._protocol,
                root_helper=None,
                use_multipath=CONF.hyperv.use_multipath_io,
                device_scan_attempts=scan_attempts,
                device_scan_interval=scan_interval,
                **self._extra_connector_args)
        return self._conn

    def connect_volume(self, connection_info):
        return self._connector.connect_volume(connection_info['data'])

    def disconnect_volume(self, connection_info):
        self._connector.disconnect_volume(connection_info['data'])

    def get_disk_resource_path(self, connection_info):
        disk_paths = self._connector.get_volume_paths(connection_info['data'])
        if not disk_paths:
            vol_id = connection_info['serial']
            err_msg = _("Could not find disk path. Volume id: %s")
            raise exception.DiskNotFound(err_msg % vol_id)

        return self._get_disk_res_path(disk_paths[0])

    def validate_host_configuration(self):
        if self._is_block_dev:
            self._check_san_policy()

    def _get_disk_res_path(self, disk_path):
        if self._is_block_dev:
            # We need the Msvm_DiskDrive resource path as this
            # will be used when the disk is attached to an instance.
            disk_number = self._diskutils.get_device_number_from_device_name(
                disk_path)
            disk_res_path = self._vmutils.get_mounted_disk_by_drive_number(
                disk_number)
        else:
            disk_res_path = disk_path

        if not disk_res_path:
            err_msg = _("Could not find an attachable disk resource path "
                        "for disk: %s") % disk_path
            raise exception.DiskNotFound(err_msg)
        return disk_res_path

    def _check_san_policy(self):
        disk_policy = self._diskutils.get_new_disk_policy()

        accepted_policies = [os_win_const.DISK_POLICY_OFFLINE_SHARED,
                             os_win_const.DISK_POLICY_OFFLINE_ALL]

        if disk_policy not in accepted_policies:
            err_msg = _("Invalid SAN policy. The SAN policy "
                        "must be set to 'Offline Shared' or 'Offline All' "
                        "in order to attach passthrough disks to instances.")
            raise exception.ValidationError(message=err_msg)

    def attach_volume(self, connection_info, instance_name,
                      disk_bus=constants.CTRL_TYPE_SCSI):
        self.validate_host_configuration()

        dev_info = self.connect_volume(connection_info)

        serial = connection_info['serial']
        disk_path = self._get_disk_res_path(dev_info['path'])
        ctrller_path, slot = self._get_disk_ctrl_and_slot(instance_name,
                                                          disk_bus)
        if self._is_block_dev:
            # We need to tag physical disk resources with the volume
            # serial number, in order to be able to retrieve them
            # during live migration.
            self._vmutils.attach_volume_to_controller(instance_name,
                                                      ctrller_path,
                                                      slot,
                                                      disk_path,
                                                      serial=serial)
        else:
            self._vmutils.attach_drive(instance_name,
                                       disk_path,
                                       ctrller_path,
                                       slot)

        self._configure_disk_metrics(disk_path)

    def _configure_disk_metrics(self, disk_path):
        if not CONF.hyperv.enable_instance_metrics_collection:
            return

        if self._is_block_dev:
            LOG.warning("Hyper-V does not support collecting metrics for "
                        "passthrough disks (e.g. iSCSI/FC).")
            return

        LOG.debug("Enabling disk metrics: %s.", disk_path)
        self._metricsutils.enable_disk_metrics_collection(
            disk_path, is_physical=self._is_block_dev)

    def detach_volume(self, connection_info, instance_name):
        if self._migrutils.planned_vm_exists(instance_name):
            LOG.warning("Instance %s is a Planned VM, cannot detach "
                        "volumes from it.", instance_name)
            return
        # Retrieving the disk path can be a time consuming operation in
        # case of passthrough disks. As such disks attachments will be
        # tagged using the volume id, we'll just use that instead.
        #
        # Note that Hyper-V does not allow us to attach the same passthrough
        # disk to multiple instances, which means that we're safe to rely
        # on this tag.
        if not self._is_block_dev:
            disk_path = self.get_disk_resource_path(connection_info)
            # In this case, we're not tagging the disks, so we want os-win
            # to use the disk path to identify the attachment.
            serial = None
        else:
            disk_path = None
            serial = connection_info['serial']

        LOG.debug("Detaching disk from instance: %(instance_name)s. "
                  "Disk path: %(disk_path)s. Disk serial tag: %(serial)s.",
                  dict(disk_path=disk_path,
                       serial=serial,
                       instance_name=instance_name))
        self._vmutils.detach_vm_disk(instance_name, disk_path,
                                     is_physical=self._is_block_dev,
                                     serial=serial)

    def _get_disk_ctrl_and_slot(self, instance_name, disk_bus):
        if disk_bus == constants.CTRL_TYPE_IDE:
            # Find the IDE controller for the vm.
            ctrller_path = self._vmutils.get_vm_ide_controller(
                instance_name, 0)
            # Attaching to the first slot
            slot = 0
        else:
            # Find the SCSI controller for the vm
            ctrller_path = self._vmutils.get_vm_scsi_controller(
                instance_name)
            slot = self._vmutils.get_free_controller_slot(ctrller_path)
        return ctrller_path, slot

    def set_disk_qos_specs(self, connection_info, disk_qos_specs):
        LOG.info("The %(protocol)s Hyper-V volume driver "
                 "does not support QoS. Ignoring QoS specs.",
                 dict(protocol=self._protocol))

    def create_snapshot(self, connection_info, instance, create_info):
        raise NotImplementedError()

    def delete_snapshot(self, connection_info, instance, delete_info):
        raise NotImplementedError()

    def get_disk_attachment_info(self, connection_info):
        if self._is_block_dev:
            disk_path = None
            serial = connection_info['serial']
        else:
            disk_path = self.get_disk_resource_path(connection_info)
            serial = None

        return self._vmutils.get_disk_attachment_info(
            disk_path,
            is_physical=self._is_block_dev,
            serial=serial)

    def extend_volume(self, connection_info):
        # We're not actually extending the volume, we're just
        # refreshing cached information about an already extended volume.
        self._connector.extend_volume(connection_info['data'])


class ISCSIVolumeDriver(BaseVolumeDriver):
    _is_block_dev = True
    _protocol = constants.STORAGE_PROTOCOL_ISCSI

    def __init__(self, *args, **kwargs):
        self._extra_connector_args = dict(
            initiator_list=CONF.hyperv.iscsi_initiator_list)

        super(ISCSIVolumeDriver, self).__init__(*args, **kwargs)


class SMBFSVolumeDriver(BaseVolumeDriver):
    _is_block_dev = False
    _protocol = constants.STORAGE_PROTOCOL_SMBFS
    _extra_connector_args = dict(local_path_for_loopback=True)

    def __init__(self):
        self._vmops_prop = None
        self._pathutils = pathutils.PathUtils()
        self._vhdutils = utilsfactory.get_vhdutils()
        super(SMBFSVolumeDriver, self).__init__()

    @property
    def _vmops(self):
        # We have to avoid a circular dependency.
        if not self._vmops_prop:
            self._vmops_prop = importutils.import_class(
                'compute_hyperv.nova.vmops.VMOps')()
        return self._vmops_prop

    def export_path_synchronized(f):
        def wrapper(inst, connection_info, *args, **kwargs):
            export_path = inst._get_export_path(connection_info)

            @utils.synchronized(export_path)
            def inner():
                return f(inst, connection_info, *args, **kwargs)
            return inner()
        return wrapper

    def _get_export_path(self, connection_info):
        return connection_info['data']['export'].replace('/', '\\')

    @export_path_synchronized
    def attach_volume(self, *args, **kwargs):
        super(SMBFSVolumeDriver, self).attach_volume(*args, **kwargs)

    @export_path_synchronized
    def disconnect_volume(self, *args, **kwargs):
        # We synchronize those operations based on the share path in order to
        # avoid the situation when a SMB share is unmounted while a volume
        # exported by it is about to be attached to an instance.
        super(SMBFSVolumeDriver, self).disconnect_volume(*args, **kwargs)

    def set_disk_qos_specs(self, connection_info, qos_specs):
        supported_qos_specs = ['total_iops_sec', 'total_bytes_sec']
        VolumeOps.validate_qos_specs(qos_specs, supported_qos_specs)

        total_bytes_sec = int(qos_specs.get('total_bytes_sec') or 0)
        total_iops_sec = int(qos_specs.get('total_iops_sec') or
                             VolumeOps.bytes_per_sec_to_iops(
                                total_bytes_sec))

        if total_iops_sec:
            disk_path = self.get_disk_resource_path(connection_info)
            self._vmutils.set_disk_qos_specs(disk_path, total_iops_sec)

    def create_snapshot(self, connection_info, instance, create_info):
        attached_path = self.get_disk_resource_path(connection_info)
        # Cinder tells us the new differencing disk file name it expects.
        # The image does not exist yet, so we'll have to create it.
        new_path = os.path.join(os.path.dirname(attached_path),
                                create_info['new_file'])
        attachment_info = self._vmutils.get_disk_attachment_info(
            attached_path, is_physical=False)
        disk_ctrl_type = attachment_info['controller_type']

        if disk_ctrl_type == constants.CTRL_TYPE_SCSI:
            self._create_snapshot_scsi(instance, attachment_info,
                                       attached_path, new_path)
        else:
            # IDE disks cannot be hotplugged.
            self._create_snapshot_ide(instance, attached_path, new_path)

        connection_info['data']['name'] = create_info['new_file']

    def _create_snapshot_ide(self, instance, attached_path, new_path):
        with self._vmops.prepare_for_volume_snapshot(instance):
            self._vhdutils.create_differencing_vhd(new_path, attached_path)
            self._vmutils.update_vm_disk_path(attached_path, new_path,
                                              is_physical=False)

    def _create_snapshot_scsi(self, instance, attachment_info,
                              attached_path, new_path):
        with self._vmops.prepare_for_volume_snapshot(instance,
                                                     allow_paused=True):
            self._vmutils.detach_vm_disk(instance.name,
                                         attached_path,
                                         is_physical=False)
            self._vhdutils.create_differencing_vhd(new_path, attached_path)
            self._vmutils.attach_drive(instance.name,
                                       new_path,
                                       attachment_info['controller_path'],
                                       attachment_info['controller_slot'])

    def delete_snapshot(self, connection_info, instance, delete_info):
        attached_path = self.get_disk_resource_path(connection_info)
        attachment_info = self._vmutils.get_disk_attachment_info(
            attached_path, is_physical=False)
        disk_ctrl_type = attachment_info['controller_type']

        base_dir = os.path.dirname(attached_path)
        file_to_merge_name = delete_info['file_to_merge']
        file_to_merge = os.path.join(base_dir, file_to_merge_name)

        allow_paused = disk_ctrl_type == constants.CTRL_TYPE_SCSI
        with self._vmops.prepare_for_volume_snapshot(
                instance,
                allow_paused=allow_paused):
            curr_state = self._vmutils.get_vm_state(instance.name)
            # We need to detach the image in order to alter the vhd chain
            # while the instance is paused.
            needs_detach = curr_state == os_win_const.HYPERV_VM_STATE_PAUSED

            if needs_detach:
                self._vmutils.detach_vm_disk(instance.name,
                                             attached_path,
                                             is_physical=False)
            new_top_img_path = self._do_delete_snapshot(attached_path,
                                                        file_to_merge)
            attachment_changed = (attached_path.lower() !=
                                  new_top_img_path.lower())

            if needs_detach:
                self._vmutils.attach_drive(instance.name,
                                           new_top_img_path,
                                           attachment_info['controller_path'],
                                           attachment_info['controller_slot'])
            elif attachment_changed:
                # When merging the latest snapshot, we have to update
                # the attachment. Luckily, although we cannot detach
                # IDE disks, we can swap them.
                self._vmutils.update_vm_disk_path(attached_path,
                                                  new_top_img_path,
                                                  is_physical=False)

            connection_info['data']['name'] = os.path.basename(
                new_top_img_path)

    def _do_delete_snapshot(self, attached_path, file_to_merge):
        parent_path = self._vhdutils.get_vhd_parent_path(file_to_merge)
        path_to_reconnect = None

        merging_top_image = attached_path.lower() == file_to_merge.lower()
        if not merging_top_image:
            path_to_reconnect = self._get_higher_image_from_chain(
                file_to_merge, attached_path)

        # We'll let Cinder delete this image. At this point, Cinder may
        # safely query it, considering that it will no longer be in-use.
        self._vhdutils.merge_vhd(file_to_merge,
                                 delete_merged_image=False)

        if path_to_reconnect:
            self._vhdutils.reconnect_parent_vhd(path_to_reconnect,
                                                parent_path)

        new_top_img_path = (parent_path if merging_top_image
                            else attached_path)
        return new_top_img_path

    def _get_higher_image_from_chain(self, vhd_path, top_vhd_path):
        # We're searching for the child image of the specified vhd.
        # We start by looking at the top image, looping through the
        # parent images.
        current_path = top_vhd_path
        parent_path = self._vhdutils.get_vhd_parent_path(current_path)
        while parent_path:
            if parent_path.lower() == vhd_path.lower():
                return current_path

            current_path = parent_path
            parent_path = self._vhdutils.get_vhd_parent_path(current_path)

        err_msg = _("Could not find image %(vhd_path)s in the chain using "
                    "top level image %(top_vhd_path)s")
        raise exception.ImageNotFound(
            err_msg % dict(vhd_path=vhd_path, top_vhd_path=top_vhd_path))


class FCVolumeDriver(BaseVolumeDriver):
    _is_block_dev = True
    _protocol = constants.STORAGE_PROTOCOL_FC


class RBDVolumeDriver(BaseVolumeDriver):
    _is_block_dev = True
    _protocol = constants.STORAGE_PROTOCOL_RBD
    _extra_connector_args = dict(do_local_attach=True)
